"""

run_tiled_inference.py

**This script is experimental, YMMV.**

Runs inference on a folder, fist splitting each image up into tiles of size
MxN (typically the native inference size of your detector), writing those
tiles out to a temporary folder, then de-duplicating the resulting detections before
merging them back into a set of detections that make sense on the original images.

This approach will likely fail to detect very large animals, so if you expect both large
and small animals (in terms of pixel size), this script is best used in
conjunction with a traditional inference pass that looks at whole images.

Currently requires temporary storage at least as large as the input data, generally
a lot more than that (depending on the overlap between adjacent tiles).  This is
inefficient, but easy to debug.

Programmatic invocation supports using YOLOv5's inference scripts (and test-time
augmentation); the command-line interface only supports standard inference right now.

"""

#%% Imports and constants

import os
import json
import tempfile
import uuid
import sys
import argparse

from multiprocessing.pool import ThreadPool
from multiprocessing.pool import Pool
from functools import partial

from tqdm import tqdm

import torch
from torchvision import ops

from megadetector.detection.run_inference_with_yolov5_val import \
    YoloInferenceOptions,run_inference_with_yolo_val
from megadetector.detection.run_detector_batch import \
    load_and_run_detector_batch,write_results_to_file,default_loaders
from megadetector.detection.run_detector import \
    try_download_known_detector, CONF_DIGITS, COORD_DIGITS
from megadetector.utils import path_utils
from megadetector.utils.ct_utils import round_float_array, round_float
from megadetector.visualization import visualization_utils as vis_utils

default_patch_overlap = 0.5
patch_jpeg_quality = 95

# This isn't NMS in the usual sense of redundant model predictions; this is being
# used to de-duplicate predictions from overlapping patches.
nms_iou_threshold = 0.45

default_tile_size = [1280,1280]

default_n_patch_extraction_workers = 1
default_pool_type = 'thread'


#%% Support functions

def get_patch_boundaries(image_size,patch_size,patch_stride=None):
    """
    Computes a list of patch starting coordinates (x,y) given an image size (w,h)
    and a stride (x,y)

    Patch size is guaranteed, but the stride may deviate to make sure all pixels are covered.
    I.e., we move by regular strides until the current patch walks off the right/bottom,
    at which point it backs up to one patch from the end.  So if your image is 15
    pixels wide and you have a stride of 10 pixels, you will get starting positions
    of 0 (from 0 to 9) and 5 (from 5 to 14).

    Args:
        image_size (tuple): size of the image you want to divide into patches, as a length-2 tuple (w,h)
        patch_size (tuple): patch size into which you want to divide an image, as a length-2 tuple (w,h)
        patch_stride (tuple or float, optional): stride between patches, as a length-2 tuple (x,y), or a
            float; if this is a float, it's interpreted as the stride relative to the patch size
            (0.1 == 10% stride).  Defaults to half the patch size.

    Returns:
        list: list of length-2 tuples, each representing the x/y start position of a patch
    """

    if patch_stride is None:
        patch_stride = (round(patch_size[0]*(1.0-default_patch_overlap)),
                        round(patch_size[1]*(1.0-default_patch_overlap)))
    elif isinstance(patch_stride,float):
        patch_stride = (round(patch_size[0]*(patch_stride)),
                        round(patch_size[1]*(patch_stride)))

    image_width = image_size[0]
    image_height = image_size[1]

    assert patch_size[0] <= image_size[0], 'Patch width {} is larger than image width {}'.format(
        patch_size[0],image_size[0])
    assert patch_size[1] <= image_size[1], 'Patch height {} is larger than image height {}'.format(
        patch_size[1],image_size[1])

    def add_patch_row(patch_start_positions,y_start):
        """
        Add one row to our list of patch start positions, i.e.
        loop over all columns.
        """

        x_start = 0; x_end = x_start + patch_size[0] - 1

        while(True):

            patch_start_positions.append([x_start,y_start])

            # If this patch put us right at the end of the last column, we're done
            if x_end == image_width - 1:
                break

            # Move one patch to the right
            x_start += patch_stride[0]
            x_end = x_start + patch_size[0] - 1

            # If this patch flows over the edge, add one more patch to cover
            # the pixels on the end, then we're done.
            if x_end > (image_width - 1):
                overshoot = (x_end - image_width) + 1
                x_start -= overshoot
                x_end = x_start + patch_size[0] - 1
                patch_start_positions.append([x_start,y_start])
                break

        # ...for each column

        return patch_start_positions

    patch_start_positions = []

    y_start = 0; y_end = y_start + patch_size[1] - 1

    while(True):

        patch_start_positions = add_patch_row(patch_start_positions,y_start)

        # If this patch put us right at the bottom of the lats row, we're done
        if y_end == image_height - 1:
            break

        # Move one patch down
        y_start += patch_stride[1]
        y_end = y_start + patch_size[1] - 1

        # If this patch flows over the bottom, add one more patch to cover
        # the pixels at the bottom, then we're done
        if y_end > (image_height - 1):
            overshoot = (y_end - image_height) + 1
            y_start -= overshoot
            y_end = y_start + patch_size[1] - 1
            patch_start_positions = add_patch_row(patch_start_positions,y_start)
            break

    # ...for each row

    for p in patch_start_positions:
        assert p[0] >= 0 and p[1] >= 0 and p[0] <= image_width and p[1] <= image_height, \
        'Patch generation error (illegal patch {})'.format(p)

    # The last patch should always end at the bottom-right of the image
    assert patch_start_positions[-1][0]+patch_size[0] == image_width, \
        'Patch generation error (last patch does not end on the right)'
    assert patch_start_positions[-1][1]+patch_size[1] == image_height, \
        'Patch generation error (last patch does not end at the bottom)'

    # All patches should be unique
    patch_start_positions_tuples = [tuple(x) for x in patch_start_positions]
    assert len(patch_start_positions_tuples) == len(set(patch_start_positions_tuples)), \
        'Patch generation error (duplicate start position)'

    return patch_start_positions

# ...get_patch_boundaries()


def patch_info_to_patch_name(image_name,patch_x_min,patch_y_min):
    """
    Gives a unique string name to an x/y coordinate, e.g. turns ("a.jpg",10,20) into
    "a.jpg_0010_0020".

    Args:
        image_name (str): image identifier
        patch_x_min (int): x coordinate
        patch_y_min (int): y coordinate

    Returns:
        str: name for this patch, e.g. "a.jpg_0010_0020"
    """
    patch_name = image_name + '_' + \
        str(patch_x_min).zfill(4) + '_' + str(patch_y_min).zfill(4)
    return patch_name


def extract_patch_from_image(im,
                             patch_xy,
                             patch_size,
                             patch_image_fn=None,
                             patch_folder=None,
                             image_name=None,
                             overwrite=True):
    """
    Extracts a patch from the provided image, and writes that patch out to a new file.

    Args:
        im (str or Image): image from which we should extract a patch, can be a filename or
            a PIL Image object.
        patch_xy (tuple): length-2 tuple of ints (x,y) representing the upper-left corner
            of the patch to extract
        patch_size (tuple): length-2 tuple of ints (w,h) representing the size of the
            patch to extract
        patch_image_fn (str, optional): image filename to write the patch to; if this is None
            the filename will be generated from [image_name] and the patch coordinates
        patch_folder (str, optional): folder in which the image lives; only used to generate
            a patch filename, so only required if [patch_image_fn] is None
        image_name (str, optional): the identifier of the source image; only used to generate
            a patch filename, so only required if [patch_image_fn] is None
        overwrite (bool, optional): whether to overwrite an existing patch image

    Returns:
        dict: a dictionary with fields xmin,xmax,ymin,ymax,patch_fn
    """

    if isinstance(im,str):
        pil_im = vis_utils.open_image(im)
    else:
        pil_im = im

    patch_x_min = patch_xy[0]
    patch_y_min = patch_xy[1]
    patch_x_max = patch_x_min + patch_size[0] - 1
    patch_y_max = patch_y_min + patch_size[1] - 1

    # PIL represents coordinates in a way that is very hard for me to get my head
    # around, such that even though the "right" and "bottom" arguments to the crop()
    # function are inclusive... well, they're not really.
    #
    # https://pillow.readthedocs.io/en/stable/handbook/concepts.html#coordinate-system
    #
    # So we add 1 to the max values.
    patch_im = pil_im.crop((patch_x_min,patch_y_min,patch_x_max+1,patch_y_max+1))
    assert patch_im.size[0] == patch_size[0]
    assert patch_im.size[1] == patch_size[1]

    if patch_image_fn is None:
        assert patch_folder is not None,\
            "If you don't supply a patch filename to extract_patch_from_image, you need to supply a folder name"
        patch_name = patch_info_to_patch_name(image_name,patch_x_min,patch_y_min)
        patch_image_fn = os.path.join(patch_folder,patch_name + '.jpg')

    if os.path.isfile(patch_image_fn) and (not overwrite):
        pass
    else:
        patch_im.save(patch_image_fn,quality=patch_jpeg_quality)

    patch_info = {}
    patch_info['xmin'] = patch_x_min
    patch_info['xmax'] = patch_x_max
    patch_info['ymin'] = patch_y_min
    patch_info['ymax'] = patch_y_max
    patch_info['patch_fn'] = patch_image_fn

    return patch_info

# ...def extract_patch_from_image(...)


def in_place_nms(md_results, iou_thres=0.45, verbose=True):
    """
    Run torch.ops.nms in-place on MD-formatted detection results.

    Args:
        md_results (dict): detection results for a list of images, in MD results format (i.e.,
            containing a list of image dicts with the key 'images', each of which has a list
            of detections with the key 'detections')
        iou_thres (float, optional): IoU threshold above which we will treat two detections as
            redundant
        verbose (bool, optional): enable additional debug console output
    """

    n_detections_before = 0
    n_detections_after = 0

    # i_image = 18; im = md_results['images'][i_image]
    for i_image,im in tqdm(enumerate(md_results['images']),total=len(md_results['images'])):

        if (im['detections'] is None) or (len(im['detections']) == 0):
            continue

        boxes = []
        scores = []

        n_detections_before += len(im['detections'])

        # det = im['detections'][0]
        for det in im['detections']:

            # Using x1/x2 notation rather than x0/x1 notation to be consistent
            # with the Torch documentation.
            x1 = det['bbox'][0]
            y1 = det['bbox'][1]
            x2 = det['bbox'][0] + det['bbox'][2]
            y2 = det['bbox'][1] + det['bbox'][3]
            box = [x1,y1,x2,y2]
            boxes.append(box)
            scores.append(det['conf'])

        # ...for each detection

        t_boxes = torch.tensor(boxes)
        t_scores = torch.tensor(scores)

        box_indices = ops.nms(t_boxes,t_scores,iou_thres).tolist()

        post_nms_detections = [im['detections'][x] for x in box_indices]

        assert len(post_nms_detections) <= len(im['detections'])

        im['detections'] = post_nms_detections

        n_detections_after += len(im['detections'])

    # ...for each image

    if verbose:
        print('NMS removed {} of {} detections'.format(
            n_detections_before-n_detections_after,
            n_detections_before))

# ...in_place_nms()


def _extract_tiles_for_image(fn_relative,image_folder,tiling_folder,patch_size,patch_stride,overwrite):
    """
    Private function to extract tiles for a single image.

    Returns a dict with fields 'patches' (see extract_patch_from_image) and 'image_fn'.

    If there is an error, 'patches' will be None and the 'error' field will contain
    failure details.  In that case, some tiles may still be generated.
    """

    fn_abs = os.path.join(image_folder,fn_relative)
    error = None
    patches = []

    image_name = path_utils.clean_filename(fn_relative,char_limit=None,force_lower=True)

    try:

        # Open the image
        im = vis_utils.open_image(fn_abs)
        image_size = [im.width,im.height]

        # Generate patch boundaries (a list of [x,y] starting points)
        patch_boundaries = get_patch_boundaries(image_size,patch_size,patch_stride)

        # Extract patches
        #
        # patch_xy = patch_boundaries[0]
        for patch_xy in patch_boundaries:

            patch_info = extract_patch_from_image(im,patch_xy,patch_size,
                                     patch_folder=tiling_folder,
                                     image_name=image_name,
                                     overwrite=overwrite)
            patch_info['source_fn'] = fn_relative
            patches.append(patch_info)

    except Exception as e:

        s = 'Patch generation error for {}: \n{}'.format(fn_relative,str(e))
        print(s)
        # patches = None
        error = s

    image_patch_info = {}
    image_patch_info['patches'] = patches
    image_patch_info['image_fn'] = fn_relative
    image_patch_info['error'] = error

    return image_patch_info


#%% Main function

def run_tiled_inference(model_file,
                        image_folder,
                        tiling_folder,
                        output_file,
                        tile_size_x=1280,
                        tile_size_y=1280,
                        tile_overlap=0.5,
                        checkpoint_path=None,
                        checkpoint_frequency=-1,
                        remove_tiles=False,
                        yolo_inference_options=None,
                        n_patch_extraction_workers=default_n_patch_extraction_workers,
                        overwrite_tiles=True,
                        image_list=None,
                        augment=False,
                        detector_options=None,
                        use_image_queue=True,
                        preprocess_on_image_queue=True,
                        loader_workers=default_loaders,
                        inference_size=None,
                        verbose=False,
                        pool_type=None):
    """
    Runs inference using [model_file] on the images in [image_folder], fist splitting each image up
    into tiles of size [tile_size_x] x [tile_size_y], writing those tiles to [tiling_folder],
    then de-duplicating the results before merging them back into a set of detections that make
    sense on the original images and writing those results to [output_file].

    [tiling_folder] can be any folder, but this function reserves the right to do whatever it wants
    within that folder, including deleting everything, so it's best if it's a new folder.
    Conceptually this folder is temporary, it's just helpful in this case to not actually
    use the system temp folder, because the tile cache may be very large, so the caller may
    want it to be on a specific drive.  If this is None, a new folder will be created in
    system temp space.

    tile_overlap is the fraction of overlap between tiles.

    Optionally removes the temporary tiles.

    if yolo_inference_options is supplied, it should be an instance of YoloInferenceOptions; in
    this case the model will be run with run_inference_with_yolov5_val.  This is typically used to
    run the model with test-time augmentation.

    Args:
        model_file (str): model filename (ending in .pt), or a well-known model name (e.g. "MDV5A")
        image_folder (str): the folder of images to proess (always recursive)
        tiling_folder (str): folder for temporary tile storage; see caveats above.  Can be None
            to use system temp space.
        output_file (str): .json file to which we should write MD-formatted results
        tile_size_x (int, optional): tile width
        tile_size_y (int, optional): tile height
        tile_overlap (float, optional): overlap between adjacent tiles, as a fraction of the
            tile size
        checkpoint_path (str, optional): checkpoint path; passed directly to run_detector_batch; see
            run_detector_batch for details
        checkpoint_frequency (int, optional): checkpoint frequency; passed directly to run_detector_batch; see
            run_detector_batch for details
        remove_tiles (bool, optional): whether to delete the tiles when we're done
        yolo_inference_options (YoloInferenceOptions, optional): if not None, will run inference with
            run_inference_with_yolov5_val.py, rather than with run_detector_batch.py, using these options
        n_patch_extraction_workers (int, optional): number of workers to use for patch extraction;
            set to <= 1 to disable parallelization
        overwrite_tiles (bool, optional): whether to overwrite image files for individual tiles if they exist
        image_list (list, optional): .json file containing a list of specific images to process.  If
            this is supplied, and the paths are absolute, [image_folder] will be ignored. If this is supplied,
            and the paths are relative, they should be relative to [image_folder]
        augment (bool, optional): apply test-time augmentation
        detector_options (dict, optional): parameters to pass to run_detector, only relevant if
            yolo_inference_options is None
        use_image_queue (bool, optional): whether to use a loader worker queue, only relevant if
            yolo_inference_options is None
        preprocess_on_image_queue (bool, optional): whether the image queue should also be responsible
            for preprocessing
        loader_workers (int, optional): number of preprocessing loader workers to use
        inference_size (int, optional): override the default inference image size, only relevant if
            yolo_inference_options is None
        verbose (bool, optional): enable additional debug output
        pool_type (str, optional): 'thread' or 'process', or None to use the default (threads)

    Returns:
        dict: MD-formatted results dictionary, identical to what's written to [output_file]
    """

    ##%% Validate arguments

    assert tile_overlap < 1 and tile_overlap >= 0, \
        'Illegal tile overlap value {}'.format(tile_overlap)

    if tile_size_x == -1:
        tile_size_x = default_tile_size[0]
    if tile_size_y == -1:
        tile_size_y = default_tile_size[1]

    patch_size = [tile_size_x,tile_size_y]
    patch_stride = (round(patch_size[0]*(1.0-tile_overlap)),
                    round(patch_size[1]*(1.0-tile_overlap)))

    if pool_type is None:
        pool_type = default_pool_type
    assert pool_type in ('thread','process'), 'Illegal pool type {}'.format(pool_type)

    if tiling_folder is None:
        tiling_folder = \
            os.path.join(tempfile.gettempdir(), 'md-tiling', str(uuid.uuid1()))
        print('Creating temporary tiling folder: {}'.format(tiling_folder))

    os.makedirs(tiling_folder,exist_ok=True)

    ##%% List files

    if image_list is None:

        print('Enumerating images in {}'.format(image_folder))
        image_files_relative = path_utils.find_images(image_folder, recursive=True, return_relative_paths=True)
        assert len(image_files_relative) > 0, 'No images found in folder {}'.format(image_folder)

    else:

        print('Loading image list from {}'.format(image_list))
        with open(image_list,'r') as f:
            image_files_relative = json.load(f)
        n_absolute_paths = 0
        for i_fn,fn in enumerate(image_files_relative):
            if os.path.isabs(fn):
                n_absolute_paths += 1
                try:
                    fn_relative = os.path.relpath(fn,image_folder)
                except ValueError:
                    raise ValueError(
                        'Illegal absolute path supplied to run_tiled_inference, {} is outside of {}'.format(
                        fn,image_folder))
                assert not fn_relative.startswith('..'), \
                    'Illegal absolute path supplied to run_tiled_inference, {} is outside of {}'.format(
                        fn,image_folder)
                image_files_relative[i_fn] = fn_relative
        if (n_absolute_paths != 0) and (n_absolute_paths != len(image_files_relative)):
            raise ValueError('Illegal file list: converted {} of {} paths to relative'.format(
            n_absolute_paths,len(image_files_relative)))

    ##%% Generate tiles

    all_image_patch_info = None

    print('Extracting patches from {} images on {} workers'.format(
        len(image_files_relative),n_patch_extraction_workers))

    n_workers = n_patch_extraction_workers

    if n_workers <= 1:

        all_image_patch_info = []

        # fn_relative = image_files_relative[0]
        for fn_relative in tqdm(image_files_relative):
            image_patch_info = \
                _extract_tiles_for_image(fn_relative,image_folder,tiling_folder,patch_size,patch_stride,
                                         overwrite=overwrite_tiles)
            all_image_patch_info.append(image_patch_info)

    else:

        pool = None
        try:
            if n_workers > len(image_files_relative):

                print('Pool of {} requested, but only {} images available, reducing pool to {}'.\
                      format(n_workers,len(image_files_relative),len(image_files_relative)))
                n_workers = len(image_files_relative)

            if pool_type == 'thread':
                pool = ThreadPool(n_workers); poolstring = 'threads'
            else:
                pool = Pool(n_workers); poolstring = 'processes'

            print('Starting patch extraction pool with {} {}'.format(n_workers,poolstring))

            all_image_patch_info = list(tqdm(pool.imap(
                    partial(_extract_tiles_for_image,
                            image_folder=image_folder,
                            tiling_folder=tiling_folder,
                            patch_size=patch_size,
                            patch_stride=patch_stride,
                            overwrite=overwrite_tiles),
                    image_files_relative),total=len(image_files_relative)))
        finally:
            if pool is not None:
                pool.close()
                pool.join()
                print('Pool closed and joined for patch extraction')

    # ...for each image

    # Write tile information to file; this is just a debugging convenience
    folder_name = path_utils.clean_filename(image_folder,force_lower=True)
    if folder_name.startswith('_'):
        folder_name = folder_name[1:]

    tile_cache_file = os.path.join(tiling_folder,folder_name + '_patch_info.json')
    with open(tile_cache_file,'w') as f:
        json.dump(all_image_patch_info,f,indent=1)

    # Keep track of patches that failed
    images_with_patch_errors = {}
    for patch_info in all_image_patch_info:
        if patch_info['error'] is not None:
            images_with_patch_errors[patch_info['image_fn']] = patch_info


    ##%% Run inference on the folder of tiles

    # When running with run_inference_with_yolov5_val, we'll pass the folder
    if yolo_inference_options is not None:

        patch_level_output_file = os.path.join(tiling_folder,folder_name + '_patch_level_results.json')

        if yolo_inference_options.model_filename is None:
            yolo_inference_options.model_filename = model_file
        else:
            assert yolo_inference_options.model_filename == model_file, \
                'Model file between yolo inference file ({}) and model file parameter ({})'.format(
                    yolo_inference_options.model_filename,model_file)

        yolo_inference_options.input_folder = tiling_folder
        yolo_inference_options.output_file = patch_level_output_file

        run_inference_with_yolo_val(yolo_inference_options)
        with open(patch_level_output_file,'r') as f:
            patch_level_results = json.load(f)

    # For standard inference, we'll pass a list of files
    else:

        patch_file_names = []
        for patch_info in all_image_patch_info:
            # If there was a patch generation error, don't run inference
            if patch_info['error'] is not None:
                assert patch_info['image_fn'] in images_with_patch_errors
                continue
            for patch in patch_info['patches']:
                patch_file_names.append(patch['patch_fn'])

        inference_results = load_and_run_detector_batch(model_file,
                                                        patch_file_names,
                                                        checkpoint_path=checkpoint_path,
                                                        checkpoint_frequency=checkpoint_frequency,
                                                        quiet=True,
                                                        augment=augment,
                                                        detector_options=detector_options,
                                                        use_image_queue=use_image_queue,
                                                        preprocess_on_image_queue=preprocess_on_image_queue,
                                                        image_size=inference_size,
                                                        verbose_output=verbose,
                                                        loader_workers=loader_workers)

        patch_level_output_file = os.path.join(tiling_folder,folder_name + '_patch_level_results.json')

        patch_level_results = write_results_to_file(inference_results,
                                                    patch_level_output_file,
                                                    relative_path_base=tiling_folder,
                                                    detector_file=model_file)

    # ...if we are/aren't using run_inference_with_yolov5_val

    ##%% Map patch-level detections back to the original images

    # Map relative paths for patches to detections
    patch_fn_relative_to_results = {}
    for im in tqdm(patch_level_results['images']):
        patch_fn_relative_to_results[im['file']] = im

    image_level_results = {}
    image_level_results['info'] = patch_level_results['info']
    image_level_results['detection_categories'] = patch_level_results['detection_categories']
    image_level_results['images'] = []

    image_fn_relative_to_patch_info = { x['image_fn']:x for x in all_image_patch_info }

    # i_image = 0; image_fn_relative = image_files_relative[i_image]
    for i_image,image_fn_relative in tqdm(enumerate(image_files_relative),
                                          total=len(image_files_relative)):

        image_fn_abs = os.path.join(image_folder,image_fn_relative)
        assert os.path.isfile(image_fn_abs)

        output_im = {}
        output_im['file'] = image_fn_relative

        # If we had a patch generation error
        if image_fn_relative in images_with_patch_errors:

            patch_info = image_fn_relative_to_patch_info[image_fn_relative]
            assert patch_info['error'] is not None

            output_im['detections'] = None
            output_im['failure'] = 'Patch generation error'
            output_im['failure_details'] = patch_info['error']
            image_level_results['images'].append(output_im)
            continue

        try:
            pil_im = vis_utils.open_image(image_fn_abs)
            image_w = pil_im.size[0]
            image_h = pil_im.size[1]

        # This would be a very unusual situation; we're reading back an image here that we already
        # (successfully) read once during patch generation.
        except Exception as e:
            print('Warning: image read error after successful patch generation for {}:\n{}'.format(
                image_fn_relative,str(e)))
            output_im['detections'] = None
            output_im['failure'] = 'Patch processing error'
            output_im['failure_details'] = str(e)
            image_level_results['images'].append(output_im)
            continue

        output_im['detections'] = []

        image_patch_info = image_fn_relative_to_patch_info[image_fn_relative]
        assert image_patch_info['patches'][0]['source_fn'] == image_fn_relative

        # Patches for this image
        patch_fn_abs_to_patch_info_this_image = {}

        for patch_info in image_patch_info['patches']:
            patch_fn_abs_to_patch_info_this_image[patch_info['patch_fn']] = patch_info

        # For each patch
        #
        # i_patch = 0; patch_fn_abs = list(patch_fn_abs_to_patch_info_this_image.keys())[i_patch]
        for i_patch,patch_fn_abs in enumerate(patch_fn_abs_to_patch_info_this_image.keys()):

            patch_fn_relative = os.path.relpath(patch_fn_abs,tiling_folder)
            patch_results = patch_fn_relative_to_results[patch_fn_relative]
            patch_info = patch_fn_abs_to_patch_info_this_image[patch_fn_abs]

            # patch_results['file'] is a relative path, and a subset of patch_info['patch_fn']
            assert patch_results['file'] in patch_info['patch_fn']

            patch_w = (patch_info['xmax'] - patch_info['xmin']) + 1
            patch_h = (patch_info['ymax'] - patch_info['ymin']) + 1
            assert patch_w == patch_size[0]
            assert patch_h == patch_size[1]

            # If there was an inference failure on one patch, report the image
            # as an inference failure
            if 'detections' not in patch_results:
                assert 'failure' in patch_results
                output_im['detections'] = None
                output_im['failure'] = patch_results['failure']
                break

            # det = patch_results['detections'][0]
            for det in patch_results['detections']:

                bbox_patch_relative = det['bbox']
                xmin_patch_relative = bbox_patch_relative[0]
                ymin_patch_relative = bbox_patch_relative[1]
                w_patch_relative = bbox_patch_relative[2]
                h_patch_relative = bbox_patch_relative[3]

                # Convert from patch-relative normalized values to image-relative absolute values
                w_pixels = w_patch_relative * patch_w
                h_pixels = h_patch_relative * patch_h
                xmin_patch_pixels = xmin_patch_relative * patch_w
                ymin_patch_pixels = ymin_patch_relative * patch_h
                xmin_image_pixels = patch_info['xmin'] + xmin_patch_pixels
                ymin_image_pixels = patch_info['ymin'] + ymin_patch_pixels

                # ...and now to image-relative normalized values
                w_image_normalized = w_pixels / image_w
                h_image_normalized = h_pixels / image_h
                xmin_image_normalized = xmin_image_pixels / image_w
                ymin_image_normalized = ymin_image_pixels / image_h

                bbox_image_normalized = [xmin_image_normalized,
                                         ymin_image_normalized,
                                         w_image_normalized,
                                         h_image_normalized]

                bbox_image_normalized = round_float_array(bbox_image_normalized,
                                                          precision=COORD_DIGITS)
                det['conf'] = round_float(det['conf'], precision=CONF_DIGITS)

                output_det = {}
                output_det['bbox'] = bbox_image_normalized
                output_det['conf'] = det['conf']
                output_det['category'] = det['category']

                output_im['detections'].append(output_det)

            # ...for each detection

        # ...for each patch

        image_level_results['images'].append(output_im)

    # ...for each image

    image_level_results_file_pre_nms = \
        os.path.join(tiling_folder,folder_name + '_image_level_results_pre_nms.json')
    with open(image_level_results_file_pre_nms,'w') as f:
        json.dump(image_level_results,f,indent=1)


    ##%% Run NMS

    in_place_nms(image_level_results,iou_thres=nms_iou_threshold)


    ##%% Write output file

    print('Saving image-level results (after NMS) to {}'.format(output_file))

    with open(output_file,'w') as f:
        json.dump(image_level_results,f,indent=1)


    ##%% Possibly remove tiles

    if remove_tiles:

        patch_file_names = []
        for im in all_image_patch_info:
            for patch in im['patches']:
                patch_file_names.append(patch['patch_fn'])

        for patch_fn_abs in patch_file_names:
            os.remove(patch_fn_abs)


    ##%% Return

    return image_level_results


#%% Interactive driver

if False:

    pass

    #%% Run tiled inference (in Python)

    model_file = os.path.expanduser('~/models/camera_traps/megadetector/md_v5.0.0/md_v5a.0.0.pt')
    image_folder = os.path.expanduser('~/data/KRU-test')
    tiling_folder = os.path.expanduser('~/tmp/tiling-test')
    output_file = os.path.expanduser('~/tmp/KRU-test-tiled.json')

    tile_size_x = 3000
    tile_size_y = 3000
    tile_overlap = 0.5
    checkpoint_path = None
    checkpoint_frequency = -1
    remove_tiles = False

    use_yolo_inference = False

    if not use_yolo_inference:

        yolo_inference_options = None

    else:

        yolo_inference_options = YoloInferenceOptions()
        yolo_inference_options.yolo_working_folder = os.path.expanduser('~/git/yolov5')

    run_tiled_inference(model_file, image_folder, tiling_folder, output_file,
                        tile_size_x=tile_size_x, tile_size_y=tile_size_y,
                        tile_overlap=tile_overlap,
                        checkpoint_path=checkpoint_path,
                        checkpoint_frequency=checkpoint_frequency,
                        remove_tiles=remove_tiles,
                        yolo_inference_options=yolo_inference_options)


    #%% Run tiled inference (generate a command)

    import os

    model_file = os.path.expanduser('~/models/camera_traps/megadetector/md_v5.0.0/md_v5a.0.0.pt')
    image_folder = os.path.expanduser('~/data/KRU-test')
    tiling_folder = os.path.expanduser('~/tmp/tiling-test')
    output_file = os.path.expanduser('~/tmp/KRU-test-tiled.json')
    tile_size = [5152,3968]
    tile_overlap = 0.8

    cmd = f'python run_tiled_inference.py {model_file} {image_folder} {tiling_folder} {output_file} ' + \
          f'--tile_overlap {tile_overlap} --no_remove_tiles --tile_size_x {tile_size[0]} --tile_size_y {tile_size[1]}'

    print(cmd)
    import clipboard; clipboard.copy(cmd)


    #%% Preview tiled inference

    from megadetector.postprocessing.postprocess_batch_results import \
        PostProcessingOptions, process_batch_results

    options = PostProcessingOptions()
    options.image_base_dir = image_folder
    options.include_almost_detections = True
    options.num_images_to_sample = None
    options.confidence_threshold = 0.2
    options.almost_detection_confidence_threshold = options.confidence_threshold - 0.05
    options.ground_truth_json_file = None
    options.separate_detections_by_category = True
    # options.sample_seed = 0

    options.parallelize_rendering = True
    options.parallelize_rendering_n_cores = 10
    options.parallelize_rendering_with_threads = False

    preview_base = os.path.join(tiling_folder,'preview')
    os.makedirs(preview_base, exist_ok=True)

    print('Processing post-RDE to {}'.format(preview_base))

    options.md_results_file = output_file
    options.output_dir = preview_base
    ppresults = process_batch_results(options)
    html_output_file = ppresults.output_html_file

    path_utils.open_file(html_output_file)


#%% Command-line driver

def main():
    """
    Command-line driver for run_tiled_inference
    """

    parser = argparse.ArgumentParser(
        description='Chop a folder of images up into tiles, run MD on the tiles, and stitch the results together')
    parser.add_argument(
        'model_file',
        help='Path to detector model file (.pb or .pt)')
    parser.add_argument(
        'image_folder',
        help='Folder containing images for inference (always recursive, unless image_list is supplied)')
    parser.add_argument(
        'tiling_folder',
        help='Temporary folder where tiles and intermediate results will be stored')
    parser.add_argument(
        'output_file',
        help='Path to output JSON results file, should end with a .json extension')
    parser.add_argument(
        '--no_remove_tiles',
        action='store_true',
        help='Tiles are removed by default; this option suppresses tile deletion')
    parser.add_argument(
        '--augment',
        action='store_true',
        help='Enable test-time augmentation')
    parser.add_argument(
        '--verbose',
        action='store_true',
        help='Enable additional debug output')
    parser.add_argument(
        '--tile_size_x',
        type=int,
        default=default_tile_size[0],
        help=('Tile width (defaults to {})'.format(default_tile_size[0])))
    parser.add_argument(
        '--tile_size_y',
        type=int,
        default=default_tile_size[1],
        help=('Tile height (defaults to {})'.format(default_tile_size[1])))
    parser.add_argument(
        '--tile_overlap',
        type=float,
        default=default_patch_overlap,
        help=('Overlap between tiles [0,1] (defaults to {})'.format(default_patch_overlap)))
    parser.add_argument(
        '--overwrite_handling',
        type=str,
        default='skip',
        help=('Behavior when the target file exists (skip/overwrite/error) (default skip)'))
    parser.add_argument(
        '--image_list',
        type=str,
        default=None,
        help=('A .json list of relative filenames (or absolute paths contained within image_folder) to include'))
    parser.add_argument(
        '--detector_options',
        type=str,
        default=None,
        help=('A list of detector options (key-value pairs)'))
    parser.add_argument(
        '--inference_size',
        type=int,
        default=None,
        help=('Run inference at a non-default size'))
    parser.add_argument(
        '--n_patch_extraction_workers',
        type=int,
        default=1,
        help=('Number of workers to use for patch extraction'))
    parser.add_argument(
        '--loader_workers',
        type=int,
        default=default_loaders,
        help=('Number of workers to use for image loading and preprocessing (0 to disable)'))

    # detector_options = parse_kvp_list(args.detector_options)

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    model_file = try_download_known_detector(args.model_file)
    assert os.path.exists(model_file), \
        'detector file {} does not exist'.format(args.model_file)

    if os.path.exists(args.output_file):
        if args.overwrite_handling == 'skip':
            print('Warning: output file {} exists, skipping'.format(args.output_file))
            return
        elif args.overwrite_handling == 'overwrite':
            print('Warning: output file {} exists, overwriting'.format(args.output_file))
        elif args.overwrite_handling == 'error':
            raise ValueError('Output file {} exists'.format(args.output_file))
        else:
            raise ValueError('Unknown output handling method {}'.format(args.overwrite_handling))


    remove_tiles = (not args.no_remove_tiles)

    use_image_queue = (args.loader_workers > 0)

    run_tiled_inference(model_file,
                        args.image_folder,
                        args.tiling_folder,
                        args.output_file,
                        tile_size_x=args.tile_size_x,
                        tile_size_y=args.tile_size_y,
                        tile_overlap=args.tile_overlap,
                        remove_tiles=remove_tiles,
                        image_list=args.image_list,
                        augment=args.augment,
                        inference_size=args.inference_size,
                        verbose=args.verbose,
                        n_patch_extraction_workers=args.n_patch_extraction_workers,
                        loader_workers=args.loader_workers,
                        use_image_queue=use_image_queue)

if __name__ == '__main__':
    main()



"""

process_video.py

Splits a video (or folder of videos) into frames, runs the frames through run_detector_batch.py,
and optionally stitches together results into a new video with detection boxes.

When possible, video processing happens in memory, without writing intermediate frames to disk.
If the caller requests that frames be saved, frames are written before processing, and the MD
results correspond to the frames that were written to disk (which simplifies, for example,
repeat detection elimination).

"""

#%% Imports

import os
import sys
import argparse

from megadetector.detection import run_detector_batch
from megadetector.utils.ct_utils import args_to_object
from megadetector.utils.ct_utils import dict_to_kvp_list, parse_kvp_list
from megadetector.detection.video_utils import _filename_to_frame_number
from megadetector.detection.video_utils import find_videos
from megadetector.detection.video_utils import run_callback_on_frames_for_folder
from megadetector.detection.run_detector import load_detector
from megadetector.postprocessing.validate_batch_results import \
        ValidateBatchResultsOptions, validate_batch_results

# Notes to self re: upcoming work on checkpointing
from megadetector.utils.ct_utils import split_list_into_fixed_size_chunks # noqa
from megadetector.detection.run_detector_batch import write_checkpoint, load_checkpoint # noqa


#%% Classes

class ProcessVideoOptions:
    """
    Options controlling the behavior of process_video()
    """

    def __init__(self):

        #: Can be a model filename (.pt or .pb) or a model name (e.g. "MDV5A")
        #:
        #: Use the string "no_detection" to indicate that you only want to extract frames,
        #: not run a model.  If you do this, you almost definitely want to set
        #: keep_extracted_frames to "True", otherwise everything in this module is a no-op.
        #: I.e., there's no reason to extract frames, do nothing with them, then delete them.
        self.model_file = 'MDV5A'

        #: Video (of folder of videos) to process
        self.input_video_file = ''

        #: .json file to which we should write results
        self.output_json_file = None

        #: If [input_video_file] is a folder, should we search for videos recursively?
        self.recursive = False

        #: Enable additional debug console output
        self.verbose = False

        #: Detections below this threshold will not be included in the output file.
        self.json_confidence_threshold = 0.005

        #: Sample every Nth frame; set to None (default) or 1 to sample every frame.  Typically
        #: we sample down to around 3 fps, so for typical 30 fps videos, frame_sample=10 is a
        #: typical value.  Mutually exclusive with [time_sample].
        self.frame_sample = None

        #: Sample frames every N seconds.  Mutually exclusive with [frame_sample]
        self.time_sample = None

        #: Run the model at this image size (don't mess with this unless you know what you're
        #: getting into)... if you just want to pass smaller frames to MD, use max_width
        self.image_size = None

        #: Enable image augmentation
        self.augment = False

        #: By default, a video with no frames (or no frames retrievable with the current parameters)
        #: is silently stored as a failure; this causes it to halt execution.
        self.exit_on_empty_video = False

        #: Detector-specific options
        self.detector_options = None

        #: Write a checkpoint file (to resume processing later) every N videos;
        #: set to -1 (default) to disable checkpointing
        self.checkpoint_frequency = -1

        #: Path to checkpoint file; None (default) for auto-generation based on output filename
        self.checkpoint_path = None

        #: Resume from a checkpoint file, or "auto" to use the most recent checkpoint in the
        #: output directory
        self.resume_from_checkpoint = None

# ...class ProcessVideoOptions


#%% Functions

def _validate_video_options(options):
    """
    Consistency checking for ProcessVideoOptions objects.
    """

    n_sampling_options_configured = 0
    if options.frame_sample is not None:
        n_sampling_options_configured += 1
    if options.time_sample is not None:
        n_sampling_options_configured += 1

    if n_sampling_options_configured > 1:
        raise ValueError('frame_sample and time_sample are mutually exclusive')

    return True


def process_videos(options):
    """
    Process a video or folder of videos through MD.

    Args:
        options (ProcessVideoOptions): all the parameters used to control this process,
            including filenames; see ProcessVideoOptions for details
    """

    ## Validate options

    # Check for incompatible options
    _validate_video_options(options)

    if options.output_json_file is None:
        video_file = options.input_video_file.replace('\\','/')
        if video_file.endswith('/'):
            video_file = video_file[:-1]
        options.output_json_file = video_file + '.json'
        print('Output file not specified, defaulting to {}'.format(
            options.output_json_file))

    assert options.output_json_file.endswith('.json'), \
        'Illegal output file {}'.format(options.output_json_file)

    if options.time_sample is not None:
        every_n_frames_param = -1 * options.time_sample
    else:
        every_n_frames_param = options.frame_sample

    if options.verbose:
        print('Processing videos from input source {}'.format(options.input_video_file))

    detector = load_detector(options.model_file,detector_options=options.detector_options)

    def frame_callback(image_np,image_id):
        return detector.generate_detections_one_image(image_np,
                                                      image_id,
                                                      detection_threshold=options.json_confidence_threshold,
                                                      augment=options.augment,
                                                      image_size=options.image_size,
                                                      verbose=options.verbose)

    """
    [md_results] will be dict with keys 'video_filenames' (list of str), 'frame_rates' (list of floats),
    'results' (list of list of dicts). 'video_filenames' will contain *relative* filenames.
    'results' is a list (one element per video) of lists (one element per frame) of whatever the
    callback returns, typically (but not necessarily) dicts in the MD results format.

    For failed videos, the frame rate will be represented by -1, and "results"
    will be a dict with at least the key "failure".
    """
    if os.path.isfile(options.input_video_file):

        video_folder = os.path.dirname(options.input_video_file)
        video_bn = os.path.basename(options.input_video_file)
        md_results = run_callback_on_frames_for_folder(input_video_folder=video_folder,
                                                       frame_callback=frame_callback,
                                                       every_n_frames=every_n_frames_param,
                                                       verbose=options.verbose,
                                                       files_to_process_relative=[video_bn],
                                                       error_on_empty_video=options.exit_on_empty_video)

    else:

        assert os.path.isdir(options.input_video_file), \
            '{} is neither a file nor a folder'.format(options.input_video_file)

        video_folder = options.input_video_file

        md_results = run_callback_on_frames_for_folder(input_video_folder=options.input_video_file,
                                                       frame_callback=frame_callback,
                                                       every_n_frames=every_n_frames_param,
                                                       verbose=options.verbose,
                                                       recursive=options.recursive,
                                                       error_on_empty_video=options.exit_on_empty_video)

    # ...whether we're processing a file or a folder

    print('Finished running MD on videos')

    video_results = md_results['results']
    video_filenames = md_results['video_filenames']
    video_frame_rates = md_results['frame_rates']

    assert len(video_results) == len(video_filenames)
    assert len(video_results) == len(video_frame_rates)

    video_list_md_format = []

    # i_video = 0; results_this_video = video_results[i_video]
    for i_video,results_this_video in enumerate(video_results):

        video_fn = video_filenames[i_video]

        im = {}
        im['file'] = video_fn
        im['frame_rate'] = video_frame_rates[i_video]
        im['frames_processed'] = []

        if isinstance(results_this_video,dict):

            assert 'failure' in results_this_video
            im['failure'] = results_this_video['failure']
            im['detections'] = None

        else:

            im['detections'] = []

            # results_one_frame = results_this_video[0]
            for results_one_frame in results_this_video:

                assert results_one_frame['file'].startswith(video_fn)

                frame_number = _filename_to_frame_number(results_one_frame['file'])

                assert frame_number not in im['frames_processed'], \
                    'Received the same frame twice for video {}'.format(im['file'])

                im['frames_processed'].append(frame_number)

                for det in results_one_frame['detections']:
                    det['frame_number'] = frame_number

                # This is a no-op if there were no above-threshold detections
                # in this frame
                im['detections'].extend(results_one_frame['detections'])

            # ...for each frame

        # ...was this a failed video?

        im['frames_processed'] = sorted(im['frames_processed'])

        video_list_md_format.append(im)

    # ...for each video

    run_detector_batch.write_results_to_file(
        video_list_md_format,
        options.output_json_file,
        relative_path_base=None,
        detector_file=options.model_file)

    validation_options = ValidateBatchResultsOptions()
    validation_options.raise_errors = True
    validation_options.check_image_existence = True
    validation_options.return_data = False
    validation_options.relative_path_base = video_folder
    validate_batch_results(options.output_json_file,options=validation_options)

# ...process_videos()


def options_to_command(options):
    """
    Convert a ProcessVideoOptions object to a corresponding command line.

    Args:
        options (ProcessVideoOptions): the options set to render as a command line

    Returns:
        str: the command line corresponding to [options]

    :meta private:
    """

    cmd = 'python process_video.py'
    cmd += ' "' + options.model_file + '"'
    cmd += ' "' + options.input_video_file + '"'

    if options.recursive:
        cmd += ' --recursive'
    if options.output_json_file is not None:
        cmd += ' --output_json_file' + ' "' + options.output_json_file + '"'
    if options.json_confidence_threshold is not None:
        cmd += ' --json_confidence_threshold ' + str(options.json_confidence_threshold)
    if options.frame_sample is not None:
        cmd += ' --frame_sample ' + str(options.frame_sample)
    if options.verbose:
        cmd += ' --verbose'
    if options.detector_options is not None and len(options.detector_options) > 0:
        cmd += ' --detector_options {}'.format(dict_to_kvp_list(options.detector_options))

    return cmd


#%% Interactive driver

if False:

    pass

    #%% Process a folder of videos

    import os
    from megadetector.detection.process_video import \
        process_videos, ProcessVideoOptions

    model_file = 'MDV5A'
    input_dir = r"G:\temp\md-test-images\video-samples"
    assert os.path.isdir(input_dir)

    output_json_file = os.path.join(input_dir,'mdv5a-video.json')

    print('Processing folder {}'.format(input_dir))

    options = ProcessVideoOptions()
    options.json_confidence_threshold = 0.05
    options.model_file = model_file
    options.input_video_file = input_dir
    options.output_json_file = output_json_file
    options.recursive = True
    # options.frame_sample = 10
    options.time_sample = 2
    options.verbose = True

    process_videos(options)


    #%% Process a single video

    import os
    from megadetector.detection.process_video import \
        process_videos, ProcessVideoOptions
    from megadetector.detection.video_utils import find_videos

    model_file = 'MDV5A'
    input_dir = r"G:\temp\md-test-images\video-samples"
    assert os.path.isdir(input_dir)
    video_fn_abs = find_videos(input_dir)[0]

    output_json_file = os.path.join(input_dir,'mdv5a-single-video.json')

    print('Processing video {}'.format(video_fn_abs))

    options = ProcessVideoOptions()
    options.json_confidence_threshold = 0.05
    options.model_file = model_file
    options.input_video_file = video_fn_abs
    options.output_json_file = output_json_file
    options.recursive = True
    # options.frame_sample = 10
    options.time_sample = 2
    options.verbose = True

    process_videos(options)


#%% Command-line driver

def main(): # noqa

    default_options = ProcessVideoOptions()

    parser = argparse.ArgumentParser(description=(
        'Run MegaDetector on each frame (or every Nth frame) in a video (or folder of videos), optionally '\
        'producing a new video with detections annotated'))

    parser.add_argument('model_file', type=str,
                        help='MegaDetector model file (.pt or .pb) or model name (e.g. "MDV5A"), '\
                             'or the string "no_detection" to run just frame extraction')

    parser.add_argument('input_video_file', type=str,
                        help='video file (or folder) to process')

    parser.add_argument('--recursive', action='store_true',
                        help='recurse into [input_video_file]; only meaningful if a folder '\
                         'is specified as input')

    parser.add_argument('--output_json_file', type=str,
                        default=None, help='.json output file, defaults to [video file].json')

    parser.add_argument('--json_confidence_threshold', type=float,
                        default=default_options.json_confidence_threshold,
                        help="don't include boxes in the .json file with confidence "\
                            'below this threshold (default {})'.format(
                                default_options.json_confidence_threshold))

    parser.add_argument('--frame_sample', type=int,
                        default=None, help='process every Nth frame (defaults to every frame), mutually exclusive '\
                            'with --time_sample.')

    parser.add_argument('--time_sample', type=float,
                        default=None, help='process frames every N seconds; this is converted to a '\
                            'frame sampling rate, so it may not be exactly the requested interval in seconds. '\
                            'mutually exclusive with --frame_sample')

    parser.add_argument('--verbose', action='store_true',
                        help='Enable additional debug output')

    parser.add_argument('--image_size',
                        type=int,
                        default=None,
                        help=('Force image resizing to a specific integer size on the long '\
                              'axis (not recommended to change this)'))

    parser.add_argument('--augment',
                        action='store_true',
                        help='Enable image augmentation')

    parser.add_argument('--exit_on_empty_video',
                        action='store_true',
                        help=('By default, videos with no retrievable frames are stored as failures; this' \
                              'causes them to halt execution'))

    parser.add_argument(
        '--detector_options',
        nargs='*',
        metavar='KEY=VALUE',
        default='',
        help='Detector-specific options, as a space-separated list of key-value pairs')

    parser.add_argument(
        '--checkpoint_frequency',
        type=int,
        default=default_options.checkpoint_frequency,
        help='Write a checkpoint file (to resume processing later) every N videos; ' + \
             'set to -1 to disable checkpointing (default {})'.format(
                 default_options.checkpoint_frequency))

    parser.add_argument(
        '--checkpoint_path',
        type=str,
        default=None,
        help='Path to checkpoint file; defaults to a file in the same directory ' + \
             'as the output file')

    parser.add_argument(
        '--resume_from_checkpoint',
        type=str,
        default=None,
        help='Resume from a specific checkpoint file, or "auto" to resume from the ' + \
             'most recent checkpoint in the output directory')

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()
    options = ProcessVideoOptions()
    args_to_object(args,options)

    options.detector_options = parse_kvp_list(args.detector_options)

    if os.path.isdir(options.input_video_file):
        process_videos(options)
    else:
        assert os.path.isfile(options.input_video_file), \
            '{} is not a valid file or folder name'.format(options.input_video_file)
        assert not options.recursive, \
            '--recursive is only meaningful when processing a folder'
        process_videos(options)

if __name__ == '__main__':
    main()



"""

classification_postprocessing.py

Functions for postprocessing species classification results, particularly:

* Smoothing results within an image (an image with 700 cows and one deer is really just 701
  cows)
* Smoothing results within a sequence (a sequence that looks like deer/deer/deer/elk/deer/deer
  is really just a deer)

"""

#%% Constants and imports

import os
import json
import copy
import pandas as pd

from collections import defaultdict
from tqdm import tqdm

from megadetector.utils.ct_utils import is_list_sorted
from megadetector.utils.ct_utils import is_empty
from megadetector.utils.ct_utils import sort_dictionary_by_value
from megadetector.utils.ct_utils import sort_dictionary_by_key
from megadetector.utils.ct_utils import invert_dictionary
from megadetector.utils.ct_utils import write_json

from megadetector.utils.wi_taxonomy_utils import clean_taxonomy_string
from megadetector.utils.wi_taxonomy_utils import taxonomy_level_index
from megadetector.utils.wi_taxonomy_utils import taxonomy_level_string_to_index

from megadetector.utils.wi_taxonomy_utils import human_prediction_string
from megadetector.utils.wi_taxonomy_utils import animal_prediction_string
from megadetector.utils.wi_taxonomy_utils import is_taxonomic_prediction_string
from megadetector.utils.wi_taxonomy_utils import blank_prediction_string # noqa


#%% Options classes

class ClassificationSmoothingOptions:
    """
    Options used to parameterize smooth_classification_results_image_level()
    and smooth_classification_results_sequence_level()
    """

    def __init__(self):

        #: How many detections do we need in a dominant category to overwrite
        #: non-dominant classifications?  This is irrelevant if
        #: max_detections_nondominant_class <= 1.
        self.min_detections_to_overwrite_secondary = 4

        #: Even if we have a dominant class, if a non-dominant class has at least
        #: this many classifications in an image, leave them alone.
        #:
        #: If this is <= 1, we won't replace non-dominant, non-other classes
        #: with the dominant class, even if there are 900 cows and 1 deer.
        self.max_detections_nondominant_class = 1

        #: How many detections do we need in a dominant category to overwrite
        #: non-dominant classifications in the same family?  If this is <= 0,
        #: we'll skip this step.  This option doesn't mean anything if
        #: max_detections_nondominant_class_same_family <= 1.
        self.min_detections_to_overwrite_secondary_same_family = 2

        #: If we have this many classifications of a nondominant category,
        #: we won't do same-family overwrites.  <= 1 means "even if there are
        #: a million deer, if there are two million moose, call all the deer
        #: moose".  This option doesn't mean anything if
        #: min_detections_to_overwrite_secondary_same_family <= 0.
        self.max_detections_nondominant_class_same_family = -1

        #: If the dominant class has at least this many classifications, overwrite
        #: "other" classifications with the dominant class
        self.min_detections_to_overwrite_other = 2

        #: Names to treat as "other" categories; can't be None, but can be empty
        #:
        #: "Other" classifications will be changed to the dominant category, regardless
        #: of confidence, as long as there are at least min_detections_to_overwrite_other
        #: examples of the dominant class.  For example, cow/other will remain unchanged,
        #: but cow/cow/other will become cow/cow/cow.
        self.other_category_names = ['other','unknown','no cv result','animal','blank','mammal']

        #: We're not even going to mess around with classifications below this threshold.
        #:
        #: We won't count them, we won't over-write them, they don't exist during the
        #: within-image smoothing step.
        self.classification_confidence_threshold = 0.5

        #: We're not even going to mess around with detections below this threshold.
        #:
        #: We won't count them, we won't over-write them, they don't exist during the
        #: within-image smoothing step.
        self.detection_confidence_threshold = 0.15

        #: If classification descriptions are present and appear to represent taxonomic
        #: information, should we propagate classifications when lower-level taxa are more
        #: common in an image?  For example, if we see "carnivore/fox/fox/deer", should
        #: we make that "fox/fox/fox/deer"?
        self.propagate_classifications_through_taxonomy = True

        #: When propagating classifications down through taxonomy levels, we have to
        #: decide whether we prefer more frequent categories or more specific categories.
        #: taxonomy_propagation_level_weight and taxonomy_propagation_count_weight
        #: balance levels against counts in this process.
        self.taxonomy_propagation_level_weight = 1.0

        #: When propagating classifications down through taxonomy levels, we have to
        #: decide whether we prefer more frequent categories or more specific categories.
        #: taxonomy_propagation_level_weight and taxonomy_propagation_count_weight
        #: balance levels against counts in this process.
        #:
        #: With a very low default value, this just breaks ties.
        self.taxonomy_propagation_count_weight = 0.01

        #: Should we record information about the state of labels prior to smoothing?
        self.add_pre_smoothing_description = True

        #: When a dict (rather than a file) is passed to either smoothing function,
        #: if this is True, we'll make a copy of the input dict before modifying.
        self.modify_in_place = False

        #: Only include these categories in the smoothing process (None to use all categories)
        self.detection_category_names_to_smooth = ['animal']

        #: Debug options
        self.break_at_image = None

        ## Populated internally

        #: Only include these categories in the smoothing process (None to use all categories)
        self._detection_category_ids_to_smooth = None


#%% Utility functions

def _results_for_sequence(images_this_sequence,filename_to_results):
    """
    Fetch MD results for every image in this sequence, based on the 'file_name' field
    """

    results_this_sequence = []
    for im in images_this_sequence:
        fn = im['file_name']
        results_this_image = filename_to_results[fn]
        assert isinstance(results_this_image,dict)
        results_this_sequence.append(results_this_image)

    return results_this_sequence


def _sort_images_by_time(images):
    """
    Returns a copy of [images], sorted by the 'datetime' field (ascending).
    """
    return sorted(images, key = lambda im: im['datetime'])


def _detection_is_relevant_for_smoothing(det,options):
    """
    Determine whether [det] has classifications that might be meaningful for smoothing.
    """

    if ('classifications' not in det) or \
        (det['conf'] < options.detection_confidence_threshold):
        return False

    # Ignore non-smoothed categories
    if (options._detection_category_ids_to_smooth is not None) and \
        (det['category'] not in options._detection_category_ids_to_smooth):
        return False

    return True


def count_detections_by_classification_category(detections,options=None):
    """
    Count the number of instances of each classification category in the detections list
    [detections] that have an above-threshold detection.  Sort results in descending
    order by count.  Returns a dict mapping category ID --> count.  If no detections
    are above threshold, returns an empty dict.

    Only processes the top classification for each detection.

    Args:
        detections (list of dict): detections list
        options (ClassificationSmoothingOptions, optional): see ClassificationSmoothingOptions

    Returns:
        dict mapping above-threshold category IDs to counts
    """

    if detections is None or len(detections) == 0:
        return {}

    if options is None:
        options = ClassificationSmoothingOptions()

    category_to_count = defaultdict(int)

    for det in detections:

        if not _detection_is_relevant_for_smoothing(det,options):
            continue

        c = det['classifications'][0]
        if c[1] >= options.classification_confidence_threshold:
            category_to_count[c[0]] += 1

    category_to_count = {k: v for k, v in sorted(category_to_count.items(),
                                                 key=lambda item: item[1],
                                                 reverse=True)}

    return category_to_count


def get_classification_description_string(category_to_count,classification_descriptions):
    """
    Return a string summarizing the image content according to [category_to_count].

    Args:
        category_to_count (dict): a dict mapping category IDs to counts
        classification_descriptions (dict): a dict mapping category IDs to description strings

    Returns:
        string: a description of this image's content, e.g. "rabbit (4), human (1)"
    """

    category_strings = []
    # category_id = next(iter(category_to_count))
    for category_id in category_to_count:
        category_description = classification_descriptions[category_id]
        tokens = category_description.split(';')
        assert len(tokens) == 7
        category_name = tokens[-1]
        if len(category_name) == 0:
            category_name = 'undefined category'
        count = category_to_count[category_id]
        category_string = '{} ({})'.format(category_name,count)
        category_strings.append(category_string)

    return ', '.join(category_strings)


def _print_counts_with_names(category_to_count,classification_descriptions):
    """
    Print a list of classification categories with counts, based in the name --> count
    dict [category_to_count]
    """

    for category_id in category_to_count:
        category_name = classification_descriptions[category_id]
        count = category_to_count[category_id]
        print('{}: {} ({})'.format(category_id,category_name,count))


def _prepare_results_for_smoothing(input_file,options):
    """
    Load results from [input_file] if necessary, prepare category descriptions
    for smoothing.  Adds pre-smoothing descriptions to every image if the options
    say we're supposed to do that.

    May modify some fields in [options].
    """

    if isinstance(input_file,str):
        with open(input_file,'r') as f:
            print('Loading results from:\n{}'.format(input_file))
            d = json.load(f)
    else:
        assert isinstance(input_file,dict)
        if options.modify_in_place:
            d = input_file
        else:
            print('modify_in_place is False, copying the input before modifying')
            d = copy.deepcopy(input_file)


    ## Category processing

    category_name_to_id = {d['classification_categories'][k]:k for k in d['classification_categories']}
    other_category_ids = []
    for s in options.other_category_names:
        if s in category_name_to_id:
            other_category_ids.append(category_name_to_id[s])

    # Possibly update the list of category IDs we should smooth
    if options.detection_category_names_to_smooth is None:
        options._detection_category_ids_to_smooth = None
    else:
        detection_category_id_to_name = d['detection_categories']
        detection_category_name_to_id = invert_dictionary(detection_category_id_to_name)
        options._detection_category_ids_to_smooth = []
        for category_name in options.detection_category_names_to_smooth:
            options._detection_category_ids_to_smooth.append(detection_category_name_to_id[category_name])

    # Before we do anything else, get rid of everything but the top classification
    # for each detection, and remove the 'classifications' field from detections with
    # no classifications.
    for im in tqdm(d['images']):

        if 'detections' not in im or im['detections'] is None or len(im['detections']) == 0:
            continue

        detections = im['detections']

        for det in detections:

            if 'classifications' not in det:
                continue
            if len(det['classifications']) == 0:
                del det['classifications']
                continue

            classification_confidence_values = [c[1] for c in det['classifications']]
            assert is_list_sorted(classification_confidence_values,reverse=True)
            det['classifications'] = [det['classifications'][0]]

        # ...for each detection in this image

    # ...for each image


    ## Clean up classification descriptions...

    # ...so we can test taxonomic relationships by substring testing.

    classification_descriptions_clean = None
    classification_descriptions = None

    if 'classification_category_descriptions' in d:
        classification_descriptions = d['classification_category_descriptions']
        classification_descriptions_single = {}
        # We use "|" to delimit multiple descriptions, just use the first
        # for smoothing.  This isn't perfect or "correct", but it's reasonable.
        for k in classification_descriptions.keys():
            v = classification_descriptions[k]
            v = v.split('|')[0]
            classification_descriptions_single[k] = v
        classification_descriptions = classification_descriptions_single
        classification_descriptions_clean = {}
        # category_id = next(iter(classification_descriptions))
        for category_id in classification_descriptions:
            classification_descriptions_clean[category_id] = \
                clean_taxonomy_string(classification_descriptions[category_id]).strip(';').lower()


    ## Optionally add pre-smoothing descriptions to every image

    if options.add_pre_smoothing_description and (classification_descriptions is not None):

        for im in tqdm(d['images']):

            if 'detections' not in im or im['detections'] is None or len(im['detections']) == 0:
                continue

            detections = im['detections']
            category_to_count = count_detections_by_classification_category(detections, options)

            im['pre_smoothing_description'] = \
                get_classification_description_string(category_to_count, classification_descriptions)


    return {
        'd':d,
        'other_category_ids':other_category_ids,
        'classification_descriptions_clean':classification_descriptions_clean,
        'classification_descriptions':classification_descriptions
    }

# ...def _prepare_results_for_smoothing(...)


def _smooth_classifications_for_list_of_detections(detections,
                                                   options,
                                                   other_category_ids,
                                                   classification_descriptions,
                                                   classification_descriptions_clean):
    """
    Smooth classifications for a list of detections, which may have come from a single
    image, or may represent an entire sequence.

    Returns None if no changes are made, else a dict.

    classification_descriptions_clean should be semicolon-delimited taxonomic strings
    from which common names and GUIDs have already been removed.

    Assumes there is only one classification per detection, i.e. that non-top classifications
    have already been remoevd.
    """

    ## Count the number of instances of each category in this image

    category_to_count = count_detections_by_classification_category(detections, options)
    # _print_counts_with_names(category_to_count,classification_descriptions)
    # get_classification_description_string(category_to_count, classification_descriptions)

    if len(category_to_count) <= 1:
        return None

    keys = list(category_to_count.keys())

    # Handle a quirky special case: if the most common category is "other" and
    # it's "tied" with the second-most-common category, swap them
    if (len(keys) > 1) and \
        (keys[0] in other_category_ids) and \
        (keys[1] not in other_category_ids) and \
        (category_to_count[keys[0]] == category_to_count[keys[1]]):
            keys[1], keys[0] = keys[0], keys[1]

    max_count = category_to_count[keys[0]]
    most_common_category = keys[0]
    del keys


    ## Debug tools

    verbose_debug_enabled = False

    if options.break_at_image is not None:
        for det in detections:
            if 'image_filename' in det and \
                det['image_filename'] == options.break_at_image:
                verbose_debug_enabled = True
                break

    if verbose_debug_enabled:
        _print_counts_with_names(category_to_count,classification_descriptions)
        # from IPython import embed; embed()


    ## Possibly change "other" classifications to the most common category

    # ...if the dominant category is not an "other" category.

    n_other_classifications_changed_this_image = 0

    # If we have at least *min_detections_to_overwrite_other* in a category that isn't
    # "other", change all "other" classifications to that category
    if (max_count >= options.min_detections_to_overwrite_other) and \
        (most_common_category not in other_category_ids):

        for det in detections:

            if not _detection_is_relevant_for_smoothing(det,options):
                continue

            assert len(det['classifications']) == 1
            c = det['classifications'][0]

            if (c[1] >= options.classification_confidence_threshold) and \
               (c[0] in other_category_ids):

                if verbose_debug_enabled:
                    print('Replacing {} with {}'.format(
                        classification_descriptions[c[0]],
                        most_common_category))

                n_other_classifications_changed_this_image += 1
                c[0] = most_common_category

            # ...if there are classifications for this detection

        # ...for each detection

    # ...if we should overwrite all "other" classifications

    if verbose_debug_enabled:
        print('Made {} other changes'.format(n_other_classifications_changed_this_image))


    ## Re-count

    category_to_count = count_detections_by_classification_category(detections, options)
    # _print_counts_with_names(category_to_count,classification_descriptions)
    keys = list(category_to_count.keys())
    max_count = category_to_count[keys[0]]
    most_common_category = keys[0]
    del keys


    ## Possibly change some non-dominant classifications to the dominant category

    process_taxonomic_rules = \
        (classification_descriptions_clean is not None) and \
        (len(classification_descriptions_clean) > 0) and \
        (len(category_to_count) > 1)

    n_detections_flipped_this_image = 0

    # Don't do this if the most common category is an "other" category, or
    # if we don't have enough of the most common category
    if (most_common_category not in other_category_ids) and \
       (max_count >= options.min_detections_to_overwrite_secondary):

        # i_det = 0; det = detections[i_det]
        for i_det,det in enumerate(detections):

            if not _detection_is_relevant_for_smoothing(det,options):
                continue

            assert len(det['classifications']) == 1
            c = det['classifications'][0]

            # Don't over-write the most common category with itself
            if c[0] == most_common_category:
                continue

            # Don't bother with below-threshold classifications
            if c[1] < options.classification_confidence_threshold:
                continue

            # If we're doing taxonomic processing, at this stage, don't turn children
            # into parents; we'll likely turn parents into children in the next stage.
            if process_taxonomic_rules:

                most_common_category_description = \
                    classification_descriptions_clean[most_common_category]

                category_id_this_classification = c[0]
                assert category_id_this_classification in category_to_count

                category_description_this_classification = \
                    classification_descriptions_clean[category_id_this_classification]

                # An empty description corresponds to the "animal" category.  We don't handle
                # "animal" here as a parent category, that would be handled in the "other smoothing"
                # step above.
                if len(category_description_this_classification) == 0:
                    continue

                most_common_category_is_parent_of_this_category = \
                    most_common_category_description in category_description_this_classification

                if most_common_category_is_parent_of_this_category:
                    continue

            # If we have fewer of this category than the most common category,
            # but not *too* many, flip it to the most common category.
            if (max_count > category_to_count[c[0]]) and \
               (category_to_count[c[0]] <= options.max_detections_nondominant_class):

                c[0] = most_common_category
                n_detections_flipped_this_image += 1

        # ...for each detection

    # ...if the dominant category is legit

    if verbose_debug_enabled:
        print('Made {} non-dominant --> dominant changes'.format(
            n_detections_flipped_this_image))


    ## Re-count

    category_to_count = count_detections_by_classification_category(detections, options)
    # _print_counts_with_names(category_to_count,classification_descriptions)
    keys = list(category_to_count.keys())
    max_count = category_to_count[keys[0]]
    most_common_category = keys[0]
    del keys


    ## Possibly collapse higher-level taxonomic predictions down to lower levels

    n_taxonomic_changes_this_image = 0

    process_taxonomic_rules = \
        (classification_descriptions_clean is not None) and \
        (len(classification_descriptions_clean) > 0) and \
        (len(category_to_count) > 1)

    if process_taxonomic_rules and options.propagate_classifications_through_taxonomy:

        # det = detections[3]
        for det in detections:

            if not _detection_is_relevant_for_smoothing(det,options):
                continue

            assert len(det['classifications']) == 1
            c = det['classifications'][0]

            # Don't bother with any classifications below the confidence threshold
            if c[1] < options.classification_confidence_threshold:
                continue

            category_id_this_classification = c[0]
            assert category_id_this_classification in category_to_count

            category_description_this_classification = \
                classification_descriptions_clean[category_id_this_classification]

            # An empty description corresponds to the "animal" category.  We don't handle
            # "animal" here as a parent category, that would be handled in the "other smoothing"
            # step above.
            if len(category_description_this_classification) == 0:
                continue

            # We may have multiple child categories to choose from; this keeps track of
            # the "best" we've seen so far.  "Best" is based on the level (species is better
            # than genus) and number.
            child_category_to_score = defaultdict(float)

            for category_id_of_candidate_child in category_to_count.keys():

                # A category is never its own child
                if category_id_of_candidate_child == category_id_this_classification:
                    continue

                # Is this candidate a child of the current classification?
                category_description_candidate_child = \
                    classification_descriptions_clean[category_id_of_candidate_child]

                # An empty description corresponds to "animal", which can never
                # be a child of another category.
                if len(category_description_candidate_child) == 0:
                    continue

                # This handles a case that doesn't come up with "pure" SpeciesNet results;
                # if two categories have different IDs but the same "clean" description, this
                # means they're different common names for the same species, which we use
                # for things like "white-tailed deer buck" and "white-tailed deer fawn".
                #
                # Currently we don't support smoothing those predictions, because it's not a
                # parent/child relationship.
                if category_description_candidate_child == \
                    category_description_this_classification:
                    continue

                # As long as we're using "clean" descriptions, parent/child taxonomic
                # relationships are defined by a substring relationship
                is_child = category_description_this_classification in \
                    category_description_candidate_child

                if not is_child:
                    continue

                # How many instances of this child category are there?
                child_category_count = category_to_count[category_id_of_candidate_child]

                # What taxonomy level is this child category defined at?
                child_category_level = taxonomy_level_index(
                    classification_descriptions[category_id_of_candidate_child])

                child_category_to_score[category_id_of_candidate_child] = \
                    child_category_level * options.taxonomy_propagation_level_weight + \
                    child_category_count * options.taxonomy_propagation_count_weight

            # ...for each category we are considering reducing this classification to

            # Did we find a category we want to change this classification to?
            if len(child_category_to_score) > 0:

                # Find the child category with the highest score
                child_category_to_score = sort_dictionary_by_value(
                    child_category_to_score,reverse=True)
                best_child_category = next(iter(child_category_to_score.keys()))

                if verbose_debug_enabled:
                    old_category_name = \
                        classification_descriptions_clean[c[0]]
                    new_category_name = \
                        classification_descriptions_clean[best_child_category]
                    print('Replacing {} with {}'.format(
                        old_category_name,new_category_name))

                c[0] = best_child_category
                n_taxonomic_changes_this_image += 1

        # ...for each detection

    # ...if we have taxonomic information available


    ## Re-count

    category_to_count = count_detections_by_classification_category(detections, options)
    # _print_counts_with_names(category_to_count,classification_descriptions)
    keys = list(category_to_count.keys())
    max_count = category_to_count[keys[0]]
    most_common_category = keys[0]
    del keys


    ## Possibly do within-family smoothing

    n_within_family_smoothing_changes = 0

    # min_detections_to_overwrite_secondary_same_family = -1
    # max_detections_nondominant_class_same_family = 1
    family_level = taxonomy_level_string_to_index('family')

    if process_taxonomic_rules:

        category_description_most_common_category = \
            classification_descriptions[most_common_category]
        most_common_category_taxonomic_level = \
            taxonomy_level_index(category_description_most_common_category)
        n_most_common_category = category_to_count[most_common_category]
        tokens = category_description_most_common_category.split(';')
        assert len(tokens) == 7
        most_common_category_family = tokens[3]
        most_common_category_genus = tokens[4]

    # Only consider remapping to genus or species level, and only when we have
    # a high enough count in the most common category
    if process_taxonomic_rules and \
        (options.min_detections_to_overwrite_secondary_same_family > 0) and \
        (most_common_category not in other_category_ids) and \
        (most_common_category_taxonomic_level > family_level) and \
        (n_most_common_category >= options.min_detections_to_overwrite_secondary_same_family):

        # det = detections[0]
        for det in detections:

            if not _detection_is_relevant_for_smoothing(det,options):
                continue

            assert len(det['classifications']) == 1
            c = det['classifications'][0]

            # Don't over-write the most common category with itself
            if c[0] == most_common_category:
                continue

            # Don't bother with below-threshold classifications
            if c[1] < options.classification_confidence_threshold:
               continue

            n_candidate_flip_category = category_to_count[c[0]]

            # Do we have too many of the non-dominant category to do this kind of swap?
            if n_candidate_flip_category > \
                options.max_detections_nondominant_class_same_family:
                continue

            # Don't flip classes when it's a tie
            if n_candidate_flip_category == n_most_common_category:
                continue

            category_description_candidate_flip = \
                classification_descriptions[c[0]]
            tokens = category_description_candidate_flip.split(';')
            assert len(tokens) == 7
            candidate_flip_category_family = tokens[3]
            candidate_flip_category_genus = tokens[4]
            candidate_flip_category_taxonomic_level = \
                taxonomy_level_index(category_description_candidate_flip)

            # Only proceed if we have valid family strings
            if (len(candidate_flip_category_family) == 0) or \
                (len(most_common_category_family) == 0):
                continue

            # Only proceed if the candidate and the most common category are in the same family
            if candidate_flip_category_family != most_common_category_family:
                continue

            # Don't flip from a species to the genus level in the same genus
            if (candidate_flip_category_genus == most_common_category_genus) and \
                (candidate_flip_category_taxonomic_level > \
                 most_common_category_taxonomic_level):
                continue

            old_category_name = classification_descriptions_clean[c[0]]
            new_category_name = classification_descriptions_clean[most_common_category]

            c[0] = most_common_category
            n_within_family_smoothing_changes += 1

        # ...for each detection

    # ...if the dominant category is legit and we have taxonomic information available

    return {'n_other_classifications_changed_this_image':n_other_classifications_changed_this_image,
            'n_detections_flipped_this_image':n_detections_flipped_this_image,
            'n_taxonomic_changes_this_image':n_taxonomic_changes_this_image,
            'n_within_family_smoothing_changes':n_within_family_smoothing_changes}

# ...def _smooth_classifications_for_list_of_detections(...)


def _smooth_single_image(im,
                         options,
                         other_category_ids,
                         classification_descriptions,
                         classification_descriptions_clean):
    """
    Smooth classifications for a single image.  Returns None if no changes are made,
    else a dict.

    classification_descriptions_clean should be semicolon-delimited taxonomic strings
    from which common names and GUIDs have already been removed.

    Assumes there is only one classification per detection, i.e. that non-top classifications
    have already been remoevd.
    """

    if 'detections' not in im or im['detections'] is None or len(im['detections']) == 0:
        return

    detections = im['detections']

    # Simplify debugging
    for det in detections:
        det['image_filename'] = im['file']

    to_return = _smooth_classifications_for_list_of_detections(detections,
        options=options,
        other_category_ids=other_category_ids,
        classification_descriptions=classification_descriptions,
        classification_descriptions_clean=classification_descriptions_clean)

    # Clean out debug information
    for det in detections:
        del det['image_filename']

    return to_return

# ...def smooth_single_image


#%% Image-level smoothing

def smooth_classification_results_image_level(input_file,output_file=None,options=None):
    """
    Smooth classifications at the image level for all results in the MD-formatted results
    file [input_file], optionally writing a new set of results to [output_file].

    This function generally expresses the notion that an image with 700 cows and one deer
    is really just 701 cows.

    Only count detections with a classification confidence threshold above
    [options.classification_confidence_threshold], which in practice means we're only
    looking at one category per detection.

    If an image has at least [options.min_detections_to_overwrite_secondary] such detections
    in the most common category, and no more than [options.max_detections_nondominant_class]
    in the second-most-common category, flip all detections to the most common
    category.

    Optionally treat some classes as particularly unreliable, typically used to overwrite an
    "other" class.

    This function also removes everything but the non-dominant classification for each detection.

    Args:
        input_file (str): MegaDetector-formatted classification results file to smooth.  Can
            also be an already-loaded results dict.
        output_file (str, optional): .json file to write smoothed results
        options (ClassificationSmoothingOptions, optional): see
          ClassificationSmoothingOptions for details.

    Returns:
        dict: MegaDetector-results-formatted dict, identical to what's written to
        [output_file] if [output_file] is not None.
    """

    ## Input validation

    if options is None:
        options = ClassificationSmoothingOptions()

    r = _prepare_results_for_smoothing(input_file, options)
    d = r['d']
    other_category_ids = r['other_category_ids']
    classification_descriptions_clean = r['classification_descriptions_clean']
    classification_descriptions = r['classification_descriptions']


    ## Smoothing

    n_other_classifications_changed = 0
    n_other_images_changed = 0
    n_taxonomic_images_changed = 0

    n_detections_flipped = 0
    n_images_changed = 0
    n_taxonomic_classification_changes = 0

    # im = d['images'][0]
    for im in tqdm(d['images']):

        r = _smooth_single_image(im,
                                 options,
                                 other_category_ids,
                                 classification_descriptions=classification_descriptions,
                                 classification_descriptions_clean=classification_descriptions_clean)

        if r is None:
            continue

        n_detections_flipped_this_image = r['n_detections_flipped_this_image']
        n_other_classifications_changed_this_image = \
            r['n_other_classifications_changed_this_image']
        n_taxonomic_changes_this_image = r['n_taxonomic_changes_this_image']

        n_detections_flipped += n_detections_flipped_this_image
        n_other_classifications_changed += n_other_classifications_changed_this_image
        n_taxonomic_classification_changes += n_taxonomic_changes_this_image

        if n_detections_flipped_this_image > 0:
            n_images_changed += 1
        if n_other_classifications_changed_this_image > 0:
            n_other_images_changed += 1
        if n_taxonomic_changes_this_image > 0:
            n_taxonomic_images_changed += 1

    # ...for each image

    print('Classification smoothing: changed {} detections on {} images'.format(
        n_detections_flipped,n_images_changed))

    print('"Other" smoothing: changed {} detections on {} images'.format(
          n_other_classifications_changed,n_other_images_changed))

    print('Taxonomic smoothing: changed {} detections on {} images'.format(
          n_taxonomic_classification_changes,n_taxonomic_images_changed))


    ## Write output

    if output_file is not None:
        print('Writing results after image-level smoothing to:\n{}'.format(output_file))
        write_json(output_file,d)

    return d

# ...def smooth_classification_results_image_level(...)


#%% Sequence-level smoothing

def smooth_classification_results_sequence_level(input_file,
                                                 cct_sequence_information,
                                                 output_file=None,
                                                 options=None):
    """
    Smooth classifications at the sequence level for all results in the MD-formatted results
    file [md_results_file], optionally writing a new set of results to [output_file].

    This function generally expresses the notion that a sequence that looks like
    deer/deer/deer/elk/deer/deer/deer/deer is really just a deer.

    Args:
        input_file (str or dict): MegaDetector-formatted classification results file to smooth
          (or already-loaded results).  If you supply a dict, it's copied by default, but
          in-place modification is supported via options.modify_in_place.
        cct_sequence_information (str, dict, or list): COCO Camera Traps file containing sequence IDs for
          each image (or an already-loaded CCT-formatted dict, or just the 'images' list from a CCT dict).
        output_file (str, optional): .json file to write smoothed results
        options (ClassificationSmoothingOptions, optional): see
          ClassificationSmoothingOptions for details.

    Returns:
        dict: MegaDetector-results-formatted dict, identical to what's written to
        [output_file] if [output_file] is not None.
    """

    ## Input validation

    if options is None:
        options = ClassificationSmoothingOptions()

    r = _prepare_results_for_smoothing(input_file, options)
    d = r['d']
    other_category_ids = r['other_category_ids']
    classification_descriptions_clean = r['classification_descriptions_clean']
    classification_descriptions = r['classification_descriptions']


    ## Make a list of images appearing in each sequence

    if isinstance(cct_sequence_information,list):
        image_info = cct_sequence_information
    elif isinstance(cct_sequence_information,str):
        print('Loading sequence information from {}'.format(cct_sequence_information))
        with open(cct_sequence_information,'r') as f:
            cct_sequence_information = json.load(f)
            image_info = cct_sequence_information['images']
    else:
        assert isinstance(cct_sequence_information,dict)
        image_info = cct_sequence_information['images']

    sequence_to_image_filenames = defaultdict(list)

    # im = image_info[0]
    for im in tqdm(image_info):
        sequence_to_image_filenames[im['seq_id']].append(im['file_name'])
    del image_info

    image_fn_to_classification_results = {}
    for im in d['images']:
        fn = im['file']
        assert fn not in image_fn_to_classification_results
        image_fn_to_classification_results[fn] = im


    ## Smoothing

    n_other_classifications_changed = 0
    n_other_sequences_changed = 0
    n_taxonomic_sequences_changed = 0
    n_within_family_sequences_changed = 0

    n_detections_flipped = 0
    n_sequences_changed = 0
    n_taxonomic_classification_changes = 0
    n_within_family_changes = 0

    # sequence_id = list(sequence_to_image_filenames.keys())[0]
    for sequence_id in sequence_to_image_filenames.keys():

        image_filenames_this_sequence = sequence_to_image_filenames[sequence_id]

        # if 'file' in image_filenames_this_sequence:
        #    from IPython import embed; embed()

        detections_this_sequence = []
        for image_filename in image_filenames_this_sequence:
            if image_filename not in image_fn_to_classification_results:
                print('Warning: {} in sequence list but not in results'.format(
                    image_filename))
                continue
            im = image_fn_to_classification_results[image_filename]
            if 'detections' not in im or im['detections'] is None:
                continue
            detections_this_sequence.extend(im['detections'])

            # Temporarily add image filenames to every detection,
            # for debugging
            for det in im['detections']:
                det['image_filename'] = im['file']

        if len(detections_this_sequence) == 0:
            continue

        r = _smooth_classifications_for_list_of_detections(
            detections=detections_this_sequence,
            options=options,
            other_category_ids=other_category_ids,
            classification_descriptions=classification_descriptions,
            classification_descriptions_clean=classification_descriptions_clean)

        if r is None:
            continue

        n_detections_flipped_this_sequence = r['n_detections_flipped_this_image']
        n_other_classifications_changed_this_sequence = \
            r['n_other_classifications_changed_this_image']
        n_taxonomic_changes_this_sequence = r['n_taxonomic_changes_this_image']
        n_within_family_changes_this_sequence = r['n_within_family_smoothing_changes']

        n_detections_flipped += n_detections_flipped_this_sequence
        n_other_classifications_changed += n_other_classifications_changed_this_sequence
        n_taxonomic_classification_changes += n_taxonomic_changes_this_sequence
        n_within_family_changes += n_within_family_changes_this_sequence

        if n_detections_flipped_this_sequence > 0:
            n_sequences_changed += 1
        if n_other_classifications_changed_this_sequence > 0:
            n_other_sequences_changed += 1
        if n_taxonomic_changes_this_sequence > 0:
            n_taxonomic_sequences_changed += 1
        if n_within_family_changes_this_sequence > 0:
            n_within_family_sequences_changed += 1

    # ...for each sequence

    print('Classification smoothing: changed {} detections in {} sequences'.format(
        n_detections_flipped,n_sequences_changed))

    print('"Other" smoothing: changed {} detections in {} sequences'.format(
          n_other_classifications_changed,n_other_sequences_changed))

    print('Taxonomic smoothing: changed {} detections in {} sequences'.format(
          n_taxonomic_classification_changes,n_taxonomic_sequences_changed))

    print('Within-family smoothing: changed {} detections in {} sequences'.format(
          n_within_family_changes,n_within_family_sequences_changed))


    ## Clean up debug information

    for im in d['images']:
        if 'detections' not in im or im['detections'] is None:
            continue
        for det in im['detections']:
            if 'image_filename' in det:
                del det['image_filename']


    ## Write output

    if output_file is not None:
        print('Writing sequence-smoothed classification results to {}'.format(
            output_file))
        write_json(output_file,d)

    return d

# ...smooth_classification_results_sequence_level(...)


def remove_classifications_from_non_animal_detections(input_file,
                                                      output_file,
                                                      animal_category_names=None):
    """
    Remove classifications from non-animal detections in a MD .json file,
    optionally writing the results to a new .json file

    Args:
        input_file (str): the MD-formatted .json file to process
        output_file (str, optional): the output file to write the modified results
        animal_category_names (list, optional): the detection category names that should
            be treated as animals (defaults to just 'animal').

    Returns:
        dict: the modified results
    """

    if animal_category_names is None:
        animal_category_names = ['animal']
    animal_category_names = set(animal_category_names)

    with open(input_file,'r') as f:
        d = json.load(f)

    category_id_to_name = d['detection_categories']

    n_classifications_removed = 0
    n_detections = 0

    # im = d['images'][0]
    for im in d['images']:

        if ('detections' not in im) or (im['detections'] is None):
            continue

        n_detections += len(im['detections'])

        for det in im['detections']:

            if 'classifications' not in det:
                continue
            category_id = det['category']
            category_name = category_id_to_name[category_id]
            if category_name not in animal_category_names:
                del det['classifications']
                n_classifications_removed += 1
                continue

        # ...for each detection

    # ...for each image

    print('Removed classifications from {} of {} detections'.format(
        n_classifications_removed,n_detections))

    if output_file is not None:
        write_json(output_file,d)

    return d

# ...def remove_classifications_from_non_animal_detections(...)


def restrict_to_taxa_list(taxa_list,
                          speciesnet_taxonomy_file,
                          input_file,
                          output_file,
                          allow_walk_down=False,
                          add_pre_filtering_description=True,
                          add_post_filtering_description=True,
                          allow_redundant_latin_names=True,
                          protected_common_names=None,
                          use_original_common_names_if_available=True,
                          verbose=True,
                          classification_threshold=None,
                          combine_redundant_categories=True):
    """
    Given a prediction file in MD .json format, likely without having had
    a geofence applied, apply a custom taxa list.

    Args:
        taxa_list (str): .csv file with at least the columns "latin" and "common"
        speciesnet_taxonomy_file (str): taxonomy filename, in the same format used for
            model release (with 7-token taxonomy entries)
        input_file (str): .json file to read, in MD format.  This can be None, in which
            case this function just validates [taxa_list].
        output_file (str): .json file to write, in MD format
        allow_walk_down (bool, optional): should we walk down the taxonomy tree
            when making mappings if a parent has only a single allowable child?
            For example, if only a single felid species is allowed, should other
            felid predictions be mapped to that species, as opposed to being mapped
            to the family?
        add_pre_filtering_description (bool, optional): should we add a new metadata
            field that summarizes each image's classifications prior to taxonomic
            restriction?
        add_post_filtering_description (bool, optional): should we add a new metadata
            field that summarizes each image's classifications after taxonomic
            restriction?
        allow_redundant_latin_names (bool, optional): if False, we'll raise an Exception
            if the same latin name appears twice in the taxonomy list; if True, we'll
            just print a warning and ignore all entries other than the first for this
            latin name
        protected_common_names (list, optional): these categories should be
            unmodified, even if they aren't used, or have the same taxonomic
            description as other categories
        use_original_common_names_if_available (bool, optional): if an "original_common"
            column is present in [taxa_list], use those common names instead of the ones
            in the taxonomy file
        verbose (bool, optional): enable additional debug output
        classification_threshold (float, optional): only relevant for the pre/post filtering
            descriptions
        combine_redundant_categories (bool, optional): whether to combine categories with the
            same common name
    """

    #%% Read target taxa list

    taxa_list_df = pd.read_csv(taxa_list)

    # Convert NaN values to empty strings
    taxa_list_df = taxa_list_df.fillna('')

    # Strip string values
    string_cols = taxa_list_df.select_dtypes(include=['object']).columns
    taxa_list_df[string_cols] = taxa_list_df[string_cols].apply(lambda col: col.str.strip())

    required_columns = ('latin','common')
    for s in required_columns:
        assert s in taxa_list_df.columns, \
            'Required column {} missing from taxonomy list file {}'.format(
                s,taxa_list)

    # Convert the "latin" and "common" columns in taxa_list_df to lowercase
    taxa_list_df['latin'] = taxa_list_df['latin'].str.lower()
    taxa_list_df['common'] = taxa_list_df['common'].str.lower()

    # Remove rows from taxa_list_df where the "latin" column is nan,
    # printing a warning for each row (with a string representation of the whole row)
    for i_row,row in taxa_list_df.iterrows():
        if len(row['latin']) == 0:
            if verbose:
                print('Warning: Skipping row with empty "latin" column in {}:\n{}\n'.format(
                    taxa_list,str(row.to_dict())))
            taxa_list_df.drop(index=i_row, inplace=True)

    # Create a dictionary mapping source Latin names to target common names
    target_latin_to_common = {}

    for i_row,row in taxa_list_df.iterrows():

        latin = row['latin']
        common = row['common']

        if use_original_common_names_if_available and \
            ('original_common' in row) and \
            (not is_empty(row['original_common'])):
                common = row['original_common'].strip().lower()

        # Valid latin names have either one token (e.g. "canidae"),
        # two tokens (e.g. "bos taurus"), or three tokens (e.g. "canis lupus familiaris")
        assert len(latin.split(' ')) in (1,2,3), \
            'Illegal binomial name {} in taxaonomy list {}'.format(
                latin,taxa_list)

        if latin in target_latin_to_common:
            error_string = \
                'scientific name {} appears multiple times in the taxonomy list'.format(
                latin)
            if allow_redundant_latin_names:
                if verbose:
                    print('Warning: {}'.format(error_string))
            else:
                raise ValueError(error_string)

        target_latin_to_common[latin] = common

    # ...for each row in the custom taxonomy list


    #%% Read taxonomy file

    with open(speciesnet_taxonomy_file,'r') as f:
        speciesnet_taxonomy_list = f.readlines()
    speciesnet_taxonomy_list = [s.strip() for s in \
                                speciesnet_taxonomy_list if len(s.strip()) > 0]

    # Maps the latin name of every taxon to the corresponding full taxon string,
    # excluding the GUID.
    #
    # For species, the key is a binomial name
    speciesnet_latin_name_to_taxon_string = {}
    speciesnet_common_name_to_taxon_string = {}

    def _insert_taxonomy_string(s):
        """
        Insert the latin  name represented by [s] into
        speciesnet_latin_name_to_taxon_string.  The latin name is
        the lowest-level non-empty token (or the genus+species if all
        tokens are populated).

        Also inserts into speciesnet_common_name_to_taxon_string.

        Sets GUIDs to a uniform value to simplify matching later.
        """

        tokens = s.split(';')
        assert len(tokens) == 7, 'Illegal taxonomy string {}'.format(s)

        # guid = tokens[0] # noqa
        tokens[0] = 'guid'
        class_name = tokens[1]
        order = tokens[2]
        family = tokens[3]
        genus = tokens[4]
        species = tokens[5]
        common_name = tokens[6]

        s = ';'.join(tokens)

        if len(class_name) == 0:
            assert common_name in ('animal','vehicle','blank'), \
                'Illegal common name {}'.format(common_name)
            return

        if len(species) > 0:
            assert all([len(s) > 0 for s in [genus,family,order]]), \
                'Higher-level taxa missing for {}: {},{},{}'.format(s,genus,family,order)
            binomial_name = genus + ' ' + species
            if binomial_name not in speciesnet_latin_name_to_taxon_string:
                speciesnet_latin_name_to_taxon_string[binomial_name] = s
        elif len(genus) > 0:
            assert all([len(s) > 0 for s in [family,order]]), \
                'Higher-level taxa missing for {}: {},{}'.format(s,family,order)
            if genus not in speciesnet_latin_name_to_taxon_string:
                speciesnet_latin_name_to_taxon_string[genus] = s
        elif len(family) > 0:
            assert len(order) > 0, \
                'Higher-level taxa missing for {}: {}'.format(s,order)
            if family not in speciesnet_latin_name_to_taxon_string:
                speciesnet_latin_name_to_taxon_string[family] = s
        elif len(order) > 0:
            if order not in speciesnet_latin_name_to_taxon_string:
                speciesnet_latin_name_to_taxon_string[order] = s
        else:
            if class_name not in speciesnet_latin_name_to_taxon_string:
                speciesnet_latin_name_to_taxon_string[class_name] = s

        if len(common_name) > 0:
            if common_name not in speciesnet_common_name_to_taxon_string:
                speciesnet_common_name_to_taxon_string[common_name] = s

    # Insert everything from the taxonomy .txt file into our taxonomy dicts
    for s in speciesnet_taxonomy_list:
        _insert_taxonomy_string(s)


    #%% Make sure all parent taxa are represented in the taxonomy

    # In theory any taxon that appears as the parent of another taxon should
    # also be in the taxonomy, but this isn't always true, so we fix it here.
    new_taxon_string_to_missing_tokens = defaultdict(list)

    # latin_name = next(iter(speciesnet_latin_name_to_taxon_string.keys()))
    for latin_name in speciesnet_latin_name_to_taxon_string.keys():

        if 'no cv result' in latin_name:
            continue

        taxon_string = speciesnet_latin_name_to_taxon_string[latin_name]
        tokens = taxon_string.split(';')

        # Don't process GUID, species, or common name
        for i_token in range(1,len(tokens)-2):

            test_token = tokens[i_token]
            if len(test_token) == 0:
                continue

            # Do we need to make up a taxon for this token?
            if test_token not in speciesnet_latin_name_to_taxon_string:
                new_tokens = [''] * 7
                new_tokens[0] = 'guid'
                for i_copy_token in range(1,i_token+1):
                    new_tokens[i_copy_token] = tokens[i_copy_token]
                new_tokens[-1] = test_token + ' species'
                assert new_tokens[-2] == '', \
                    'Illegal taxonomy string {}'.format(taxon_string)
                new_taxon_string = ';'.join(new_tokens)
                # assert new_taxon_string not in new_taxon_strings
                new_taxon_string_to_missing_tokens[new_taxon_string].append(test_token)

        # ...for each token

    # ...for each taxon

    new_taxon_string_to_missing_tokens = \
        sort_dictionary_by_key(new_taxon_string_to_missing_tokens)

    if verbose:

        print(f'Found {len(new_taxon_string_to_missing_tokens)} taxa that need to be inserted to ' + \
              'make the taxonomy valid, showing only mammals and birds here:\n')

        for taxon_string in new_taxon_string_to_missing_tokens:
            if 'mammalia' not in taxon_string and 'aves' not in taxon_string:
                continue
            missing_taxa = ','.join(new_taxon_string_to_missing_tokens[taxon_string])
            print('{} ({})'.format(taxon_string,missing_taxa))

        print('')

    for new_taxon_string in new_taxon_string_to_missing_tokens:
        _insert_taxonomy_string(new_taxon_string)

    del new_taxon_string_to_missing_tokens


    #%% Store custom common name mappings based on the taxonomy list

    speciesnet_latin_name_to_output_common_name = {}

    # latin_name = next(iter(speciesnet_latin_name_to_taxon_string.keys()))
    for latin_name in speciesnet_latin_name_to_taxon_string.keys():

        if latin_name in target_latin_to_common:
            speciesnet_latin_name_to_output_common_name[latin_name] = \
                target_latin_to_common[latin_name]


    #%% Make sure all taxa on the allow-list are in the taxonomy

    n_failed_mappings = 0

    for target_taxon_latin_name in target_latin_to_common.keys():
        if target_taxon_latin_name not in speciesnet_latin_name_to_taxon_string:
            common_name = target_latin_to_common[target_taxon_latin_name]
            s = '{} ({}) not in speciesnet taxonomy'.format(
                target_taxon_latin_name,common_name)
            if common_name in speciesnet_common_name_to_taxon_string:
                s += ' (common name maps to {})'.format(
                    speciesnet_common_name_to_taxon_string[common_name])
            print(s)
            n_failed_mappings += 1

    if n_failed_mappings > 0:
        raise ValueError('Cannot continue with taxonomic restriction')


    #%% For the allow-list, map each parent taxon to a set of allowable child taxa

    # Maps parent names to all allowed child names, or None if this is the
    # lowest-level allowable taxon on this path
    allowed_parent_taxon_to_child_taxa = defaultdict(set)

    # latin_name = next(iter(target_latin_to_common.keys()))
    for latin_name in target_latin_to_common:

        taxon_string = speciesnet_latin_name_to_taxon_string[latin_name]
        tokens = taxon_string.split(';')
        assert len(tokens) == 7, \
            'Illegal taxonomy string {}'.format(taxon_string)

        # Remove GUID and common mame
        #
        # This is now always class/order/family/genus/species
        tokens = tokens[1:-1]

        child_taxon = None

        # If this is a species
        if len(tokens[-1]) > 0:
            binomial_name = tokens[-2] + ' ' + tokens[-1]
            assert binomial_name == latin_name, \
                'Binomial/latin mismatch: {} vs {}'.format(binomial_name,latin_name)
            # If this already exists, it should only allow "None"
            if binomial_name in allowed_parent_taxon_to_child_taxa:
                assert len(allowed_parent_taxon_to_child_taxa[binomial_name]) == 1, \
                    'Species-level entry {} has multiple children'.format(binomial_name)
                assert None in allowed_parent_taxon_to_child_taxa[binomial_name], \
                    'Species-level entry {} has non-None children'.format(binomial_name)
            allowed_parent_taxon_to_child_taxa[binomial_name].add(None)
            child_taxon = binomial_name

        # The first level that can ever be a parent taxon is the genus level
        parent_token_index = len(tokens) - 2

        # Walk up from genus to family
        while(parent_token_index >= 0):

            # "None" is our leaf node marker, we should never have ''
            if child_taxon is not None:
                assert len(child_taxon) > 0

            parent_taxon = tokens[parent_token_index]

            # Don't create entries for blank taxa
            if (len(parent_taxon) > 0):

                create_child = True

                # This is the lowest-level taxon in this entry
                if (child_taxon is None):

                    # ...but we don't want to remove existing children from any parents
                    if (parent_taxon in allowed_parent_taxon_to_child_taxa) and \
                       (len(allowed_parent_taxon_to_child_taxa[parent_taxon]) > 0):
                        if verbose:
                            existing_children_string = str(allowed_parent_taxon_to_child_taxa[parent_taxon])
                            print('Not creating empty child for parent {} (already has children {})'.format(
                                parent_taxon,existing_children_string))
                        create_child = False

                # If we're adding a new child entry, clear out any leaf node markers
                else:

                    if (parent_taxon in allowed_parent_taxon_to_child_taxa) and \
                       (None in allowed_parent_taxon_to_child_taxa[parent_taxon]):

                        assert len(allowed_parent_taxon_to_child_taxa[parent_taxon]) == 1, \
                            'Illlegal parent/child configuration'

                        if verbose:
                            print('Un-marking parent {} as a leaf node because of child {}'.format(
                                parent_taxon,child_taxon))

                        allowed_parent_taxon_to_child_taxa[parent_taxon] = set()

                if create_child:
                    allowed_parent_taxon_to_child_taxa[parent_taxon].add(child_taxon)

                # If we haven't hit a non-empty taxon yet, don't update "child_taxon"
                assert len(parent_taxon) > 0
                child_taxon = parent_taxon

            # ...if we have a non-empty taxon

            parent_token_index -= 1

        # ...for each taxonomic level

    # ...for each allowed latin name

    allowed_parent_taxon_to_child_taxa = \
        sort_dictionary_by_key(allowed_parent_taxon_to_child_taxa)

    for parent_taxon in allowed_parent_taxon_to_child_taxa:
        # "None" should only ever appear alone; this marks a leaf node with no children
        if None in allowed_parent_taxon_to_child_taxa[parent_taxon]:
            assert len(allowed_parent_taxon_to_child_taxa[parent_taxon]) == 1, \
                '"None" should only appear alone in a child taxon list'


    #%% If we were just validating the custom taxa file, we're done

    if input_file is None:
        print('Finished validating custom taxonomy list')
        return


    #%% Map all predictions that exist in this dataset...

    # ...to the prediction we should generate.

    with open(input_file,'r') as f:
        input_data = json.load(f)

    input_category_id_to_common_name = input_data['classification_categories'] #noqa
    input_category_id_to_taxonomy_string = \
        input_data['classification_category_descriptions']

    input_category_id_to_output_taxon_string = {}

    # input_category_id = next(iter(input_category_id_to_taxonomy_string.keys()))
    for input_category_id in input_category_id_to_taxonomy_string.keys():

        input_taxon_string = input_category_id_to_taxonomy_string[input_category_id]
        input_taxon_tokens = input_taxon_string.split(';')
        assert len(input_taxon_tokens) == 7, \
            'Illegal taxonomy string: {}'.format(input_taxon_string)

        # Don't mess with blank/no-cv-result/human (or "animal", which is really "unknown")
        if (not is_taxonomic_prediction_string(input_taxon_string)) or \
           (input_taxon_string == human_prediction_string):
            if verbose:
                print('Not messing with non-taxonomic category {}'.format(input_taxon_string))
            input_category_id_to_output_taxon_string[input_category_id] = \
                input_taxon_string
            continue

        # Don't mess with protected categories
        common_name = input_taxon_tokens[-1]

        if (protected_common_names is not None) and \
            (common_name in protected_common_names):
            if verbose:
                print('Not messing with protected category {}:\n{}'.format(
                    common_name,input_taxon_string))
            input_category_id_to_output_taxon_string[input_category_id] = \
                input_taxon_string
            continue

        # Remove GUID and common mame

        # This is always class/order/family/genus/species
        input_taxon_tokens = input_taxon_tokens[1:-1]
        assert len(input_taxon_tokens) == 5

        # Start at the species level (the last element in input_taxon_tokens),
        # and see whether each taxon is allowed
        test_index = len(input_taxon_tokens) - 1
        target_taxon = None

        while((test_index >= 0) and (target_taxon is None)):

            # Species are represented as binomial names, i.e. when test_index is 4,
            # test_taxon_name will have two tokens (e.g. "canis lupus"), otherwise
            # test_taxon_name will have one token (e.g. "canis", or "aves")
            if (test_index == (len(input_taxon_tokens) - 1)) and \
                (len(input_taxon_tokens[-1]) > 0):
                test_taxon_name = \
                    input_taxon_tokens[-2] + ' ' + input_taxon_tokens[-1]
            else:
                test_taxon_name = input_taxon_tokens[test_index]

            # If we haven't yet found the level at which this taxon is non-empty,
            # keep going up
            if len(test_taxon_name) == 0:
                test_index -= 1
                continue

            assert test_taxon_name in speciesnet_latin_name_to_taxon_string, \
                '{} not found in taxonomy table'.format(test_taxon_name)

            # Is this taxon allowed according to the custom species list?
            if test_taxon_name in allowed_parent_taxon_to_child_taxa:

                allowed_child_taxa = allowed_parent_taxon_to_child_taxa[test_taxon_name]
                assert allowed_child_taxa is not None, \
                    'allowed_child_taxa should not be None: {}'.format(test_taxon_name)

                # If this is the lowest-level allowable token or there is not a
                # unique child, don't walk any further, even if walking down
                # is enabled.
                if None in allowed_child_taxa:
                    assert len(allowed_child_taxa) == 1, \
                        '"None" should not be listed as a child taxa with other child taxa'

                if (None in allowed_child_taxa) or (len(allowed_child_taxa) > 1):
                    target_taxon = test_taxon_name
                elif not allow_walk_down:
                    target_taxon = test_taxon_name
                else:
                    candidate_taxon = None
                    # If there's a unique child, walk back *down* the allowable
                    # taxa until we run out of unique children
                    while ((next(iter(allowed_child_taxa)) is not None) and \
                          (len(allowed_child_taxa) == 1)):
                        candidate_taxon = next(iter(allowed_child_taxa))
                        assert candidate_taxon in allowed_parent_taxon_to_child_taxa, \
                            '{} should be a subset of {}'.format(
                                candidate_taxon,allowed_parent_taxon_to_child_taxa)
                        assert candidate_taxon in speciesnet_latin_name_to_taxon_string, \
                            '{} should be a subset of {}'.format(
                                candidate_taxon,speciesnet_latin_name_to_taxon_string)
                        allowed_child_taxa = \
                            allowed_parent_taxon_to_child_taxa[candidate_taxon]
                    assert candidate_taxon is not None
                    target_taxon = candidate_taxon

            # ...if this is an allowed taxon

            test_index -= 1

        # ...for each token

        if target_taxon is None:
            output_taxon_string = animal_prediction_string
        else:
            output_taxon_string = speciesnet_latin_name_to_taxon_string[target_taxon]
        input_category_id_to_output_taxon_string[input_category_id] = output_taxon_string

    # ...for each category (mapping input category IDs to output taxon strings)

    if verbose:
        for input_category_id in input_category_id_to_output_taxon_string:
            common_name = input_category_id_to_common_name[input_category_id]
            output_taxon_string = input_category_id_to_output_taxon_string[input_category_id]
            print('Mapping {} ({}) to\n{}\n'.format(
                input_category_id,
                common_name,
                output_taxon_string))


    #%% Map input category IDs to output category IDs

    speciesnet_taxon_string_to_latin_name = \
        invert_dictionary(speciesnet_latin_name_to_taxon_string)

    input_category_id_to_output_category_id = {}
    output_taxon_string_to_category_id = {}
    output_category_id_to_common_name = {}

    substitutions_printed = set()

    for input_category_id in input_category_id_to_output_taxon_string:

        output_taxon_string = \
            input_category_id_to_output_taxon_string[input_category_id]

        output_common_name = output_taxon_string.split(';')[-1]

        # Possibly substitute a custom common name
        if output_taxon_string in speciesnet_taxon_string_to_latin_name:

            speciesnet_latin_name = speciesnet_taxon_string_to_latin_name[output_taxon_string]

            if speciesnet_latin_name in speciesnet_latin_name_to_output_common_name:
                custom_common_name = speciesnet_latin_name_to_output_common_name[speciesnet_latin_name]
                if custom_common_name != output_common_name:
                    if verbose:
                        substitution_name = custom_common_name + ' --> ' + output_common_name
                        if substitution_name not in substitutions_printed:
                            substitutions_printed.add(substitution_name)
                            print('Substituting common name {} for {}'.format(
                                custom_common_name,output_common_name))
                    output_common_name = custom_common_name

        # Do we need to create a new output category?
        if output_taxon_string not in output_taxon_string_to_category_id:
            output_category_id = str(len(output_taxon_string_to_category_id))
            output_taxon_string_to_category_id[output_taxon_string] = \
                output_category_id
            output_category_id_to_common_name[output_category_id] = \
                output_common_name
        else:
            output_category_id = \
                output_taxon_string_to_category_id[output_taxon_string]

        input_category_id_to_output_category_id[input_category_id] = \
            output_category_id

        # Sometimes-useful debug printouts
        if False:
            original_common_name = \
              input_category_id_to_common_name[input_category_id]
            original_taxon_string = \
                input_category_id_to_taxonomy_string[input_category_id]
            print('Mapping {} ({}) to:\n{} ({})\n'.format(
                original_common_name,original_taxon_string,
                output_common_name,output_taxon_string))

    # ...for each category (mapping input category IDs to output category IDs)


    #%% Remap all category labels

    assert len(set(output_taxon_string_to_category_id.keys())) == \
           len(set(output_taxon_string_to_category_id.values())), \
           'Category ID/value non-uniqueness error'

    output_category_id_to_taxon_string = \
        invert_dictionary(output_taxon_string_to_category_id)

    with open(input_file,'r') as f:
        output_data = json.load(f)

    classification_descriptions = None
    if 'classification_category_descriptions' in output_data:
        classification_descriptions = output_data['classification_category_descriptions']

    for im in tqdm(output_data['images']):

        if 'detections' not in im or im['detections'] is None:
            continue

        description_options = ClassificationSmoothingOptions()
        if classification_threshold is not None:
            description_options.classification_confidence_threshold = classification_threshold

        # Possibly prepare a pre-filtering description
        pre_filtering_description = None
        if classification_descriptions is not None and add_pre_filtering_description:
            category_to_count = \
                count_detections_by_classification_category(im['detections'],
                                                            options=description_options)
            pre_filtering_description = \
                get_classification_description_string(category_to_count,classification_descriptions)
            im['pre_filtering_description'] = pre_filtering_description

        for det in im['detections']:
            if 'classifications' in det:
                for classification in det['classifications']:
                    classification[0] = \
                        input_category_id_to_output_category_id[classification[0]]

        if classification_descriptions is not None and add_post_filtering_description:
            category_to_count = \
                count_detections_by_classification_category(im['detections'],
                                                            options=description_options)
            post_filtering_description = \
                get_classification_description_string(category_to_count,output_category_id_to_taxon_string)
            im['post_filtering_description'] = post_filtering_description

    # ...for each image

    output_data['classification_categories'] = output_category_id_to_common_name
    output_data['classification_category_descriptions'] = \
        output_category_id_to_taxon_string


    #%% Write output

    write_json(output_file,output_data)

    if combine_redundant_categories:
        _ = combine_redundant_classification_categories(input_file=output_file,
                                                        output_file=output_file)

# ...def restrict_to_taxa_list(...)


def combine_redundant_classification_categories(input_file,
                                                output_file=None,
                                                classification_threshold=0.5):
    """
    Args:
        input_file (str): .json file to read, in MD format
        output_file (str): .json file to write, in MD format
        classification_threshold (float, optional): only used when sorting
            descriptions by count

    Returns:
        dict: remapped MD-formatted dict
    """

    ##%% Read input file and list categories

    assert os.path.isfile(input_file), \
        'Input file {} not found'.format(input_file)

    with open(input_file,'r') as f:
        d = json.load(f)

    input_category_name_to_ids = defaultdict(list)

    for category_id in d['classification_categories']:
        category_name = d['classification_categories'][category_id]
        input_category_name_to_ids[category_name].append(category_id)


    ##%% Return early if there are no redundant categories

    # What's the largest number of IDs associated with a single category name?
    max_count = 0
    for category_name in input_category_name_to_ids:
        c = len(input_category_name_to_ids[category_name])
        if c > max_count:
            max_count = c

    if max_count == 1:
        if output_file is not None:
            print('No redundant categories, writing data unmodified to {}'.format(
                output_file))
            write_json(output_file,d)
        return d


    ##%% Map input category IDs to output category IDs

    input_category_id_to_output_category_id = {}

    for i_category,category_name in enumerate(input_category_name_to_ids):
        output_category_id = str(i_category)
        for input_category_id in input_category_name_to_ids[category_name]:
            input_category_id_to_output_category_id[input_category_id] = \
                output_category_id

    n_input_categories = len(d['classification_categories'])
    n_output_categories = len(input_category_name_to_ids)
    assert n_output_categories < n_input_categories
    print('Removing {} redundant categories'.format(
        n_input_categories - n_output_categories))


    ##%% Create a new category dict

    output_category_name_to_id = {}

    for input_category_id in input_category_id_to_output_category_id:
        category_name = d['classification_categories'][input_category_id]
        output_category_id = input_category_id_to_output_category_id[input_category_id]
        if category_name in output_category_name_to_id:
            assert output_category_name_to_id[category_name] == output_category_id
        else:
            output_category_name_to_id[category_name] = output_category_id


    ##%% Create new classification category descriptions

    if 'classification_category_descriptions' in d:

        assert len(d['classification_category_descriptions']) == \
               len(d['classification_categories'])

        # Sort descriptions by count overall, so we can sort by description within categories later
        description_to_count = {}
        for description in d['classification_category_descriptions'].values():
            description_to_count[description] = 0

        for im in d['images']:
            if 'detections' not in im or im['detections'] is None:
                continue
            for det in im['detections']:
                if 'classifications' not in det or det['classifications'] is None:
                    continue
                conf = det['classifications'][0][1]
                if conf < classification_threshold:
                    continue
                input_category_id = det['classifications'][0][0]
                intput_category_description = d['classification_category_descriptions'][input_category_id]
                description_to_count[intput_category_description] += 1
            # ...for each detection
        # ...for each image

        # Sorting here is just a debug convenience
        description_to_count = sort_dictionary_by_value(description_to_count,
                                                        reverse=True)

        # Create descriptions for the output categories
        output_category_id_to_descriptions = defaultdict(list)

        for input_category_id in input_category_id_to_output_category_id:
            output_category_id = input_category_id_to_output_category_id[input_category_id]
            description = d['classification_category_descriptions'][input_category_id]
            output_category_id_to_descriptions[output_category_id].append(description)

        output_classification_category_descriptions = {}

        for output_category_id in output_category_id_to_descriptions:
            descriptions = output_category_id_to_descriptions[output_category_id]
            # If this output category combines multiple input categories
            if len(descriptions) > 1:
                # Sort "descriptions" in descending order by the corresponding values
                # in description_to_count
                descriptions.sort(key=lambda x: description_to_count[x], reverse=True)
            output_classification_category_descriptions[output_category_id] = \
                '|'.join(descriptions)
        # ...for each category

        d['classification_category_descriptions'] = output_classification_category_descriptions

    # ...if we have to manage descriptions

    d['classification_categories'] = invert_dictionary(output_category_name_to_id)

    # Remap classifications
    for im in d['images']:
        if 'detections' not in im or im['detections'] is None:
            continue
        for det in im['detections']:
            if 'classifications' not in det or det['classifications'] is None:
                continue
            for i_class in range(0,len(det['classifications'])):
                input_category_id = det['classifications'][i_class][0]
                output_category_id = \
                    input_category_id_to_output_category_id[input_category_id]
                det['classifications'][i_class][0] = output_category_id
            # ...for each classification
        # ...for each detection
    # ...for each image

    if output_file is not None:
        write_json(output_file,d)

    return d

# ...def combine_redundant_classification_categories(...)



#
# xmp_integration.py
#
# Tools for loading MegaDetector batch API results into XMP metadata, specifically
# for consumption in digiKam:
#
# https://cran.r-project.org/web/packages/camtrapR/vignettes/camtrapr2.html
#
    
#%% Imports and constants

import argparse
import tkinter
from tkinter import ttk, messagebox, filedialog

import inspect
import os
import sys
import json
import pyexiv2 # type: ignore
import ntpath
import threading
import traceback

from tqdm import tqdm
from multiprocessing import Pool
from multiprocessing.pool import ThreadPool
from functools import partial

category_mapping = {'person': 'Human', 'animal': 'Animal', 'vehicle': 'Vehicle'}


#%% Class definitions

class xmp_gui:
    
    root = None
    textarea_min_threshold = None
    textarea_status = None
    textarea_remove_path = None
    textarea_rename_conf = None
    textarea_rename_cats = None
    num_threads = 1
    
class xmp_integration_options:
    
    # Folder where images are stored
    image_folder = None
    
    # .json file containing MegaDetector output
    input_file = None
    
    # String to remove from all path names, typically representing a 
    # prefix that was added during MegaDetector processing
    remove_path = None
    
    # Optionally *rename* (not copy) all images that have no detections
    # above [rename_conf] for the categories in rename_cats from x.jpg to
    # x.check.jpg
    rename_conf = None
    
    # Comma-deleted list of category names (or "all") to apply the rename_conf
    # behavior to.
    rename_cats = None
    
    # Minimum detection threshold (applies to all classes, defaults to None,
    # i.e. 0.0
    min_threshold = None
    num_threads = 1
    xmp_gui = None
    
    
#%% Functions
  
def write_status(options,s):
    
    if options.xmp_gui is None:
        return
    options.xmp_gui.textarea_status.configure(state="normal")
    options.xmp_gui.textarea_status.insert(tkinter.END, s + '\n')
    options.xmp_gui.textarea_status.configure(state="disabled")

    
n_images_processed = 0
    
def update_xmp_metadata(categories, options, rename_cats, n_images, image):
    """
    Update the XMP metadata for a single image
    """
    
    # Relative image path
    filename = ''
    
    # Absolute image path
    img_path = ''
    
    global n_images_processed
    
    try:
        
        filename = image['file']
        if options.remove_path != None and len(options.remove_path) > 0:
            filename = filename.replace(options.remove_path, '')
        img_path = os.path.join(options.image_folder, filename)
        assert os.path.isfile(img_path), 'Image {} not found'.format(img_path)
        
        # List of categories to write to XMP metadata
        image_categories = []
        
        # Categories with above-threshold detections present for
        # this image
        original_image_cats = []
        
        # Maximum confidence for each category
        original_image_cats_conf = {}
        
        for detection in image['detections']:
            
            cat = category_mapping[categories[detection['category']]]
            
            # Have we already added this to the list of categories to
            # write out to this image?
            if cat not in image_categories:
                
                # If we're supposed to compare to a threshold...
                if len(options.min_threshold) > 0 and \
                    options.min_threshold != None:
                    if float(detection['conf']) > float(options.min_threshold):
                        image_categories.append(cat)
                        original_image_cats.append(
                            categories[detection['category']])
                        
                # Else we treat *any* detection as valid...
                else:
                    image_categories.append(cat)
                    original_image_cats.append(categories[detection['category']])

            # Keep track of the highest-confidence detection for this class                
            if options.min_threshold != None and \
                len(options.min_threshold) > 0 and \
                    detection['conf'] > \
                        original_image_cats_conf.get(
                            categories[detection['category']], 0):
                            
                original_image_cats_conf[categories[detection['category']]] = \
                    detection['conf']
                    
        img = pyexiv2.Image(r'{0}'.format(img_path))
        img.modify_xmp({'Xmp.lr.hierarchicalSubject': image_categories})
        
        # If we're doing the rename/.check behavior...
        if not (options.rename_conf is None and options.rename_cats is None):
            
            matching_cats = set(rename_cats).intersection(set(original_image_cats))
            is_conf_low = False
            if options.min_threshold != None and len(options.min_threshold) > 0:
                for matching_cat in matching_cats:
                    if original_image_cats_conf[matching_cat] < float(options.rename_conf):
                        is_conf_low = True
            if options.min_threshold != None and \
                len(options.min_threshold) > 0 and \
                    len(image['detections']) == 0 or \
                (len(options.rename_conf) > 0 and \
                is_conf_low is True and \
                    len(matching_cats) > 0):
                        
                parent_folder = os.path.dirname(img_path)
                file_name = ntpath.basename(img_path)
                manual_file_name = file_name.split('.')[0]+'_check' + '.' + file_name.split('.')[1]
                os.rename(img_path, os.path.join(parent_folder, manual_file_name))
                
        if options.xmp_gui is not None:
            
            n_images_processed += 1
            percentage = round((n_images_processed)/n_images*100)
            options.xmp_gui.progress_bar['value'] = percentage
            options.xmp_gui.root.update_idletasks()
            options.xmp_gui.style.configure('text.Horizontal.Tprogress_bar',
                            text='{:g} %'.format(percentage))
                
    except Exception as e:
    
        s = 'Error processing image {}: {}'.format(filename,str(e))
        print(s)
        traceback.print_exc()
        write_status(options,s)
        
        if False:
            
            # Legacy code to rename files where XMP writing failed
            parent_folder = os.path.dirname(img_path)
            file_name = ntpath.basename(img_path)        
            failed_file_name = file_name.split('.')[0]+'_failed' + '.' + file_name.split('.')[1]
            os.rename(img_path, os.path.join(
                parent_folder, failed_file_name))


def process_input_data(options):
    """
    Main function to loop over images and modify XMP data
    """
    
    if options.xmp_gui is not None:
        
        if (options.image_folder is None) or (len(options.image_folder) == 0):
            tkinter.messagebox.showerror(title='Error', message='Image folder is not selected')
            sys.exit()
        if (options.input_file is None) or (len(options.input_file) == 0):
            tkinter.messagebox.showerror(
                title='Error', message='No MegaDetector .json file selected')
            sys.exit()
        options.remove_path = options.xmp_gui.textarea_remove_path.get()
        options.rename_conf = options.xmp_gui.textarea_rename_conf.get()
        options.rename_cats = options.xmp_gui.textarea_rename_cats.get()
        options.num_threads = options.xmp_gui.textarea_num_threads.get()
        options.min_threshold = options.xmp_gui.textarea_min_threshold.get()
            
    try:
        
        with open(options.input_file, 'r') as f:
            data = f.read()
    
        data = json.loads(data)
        categories = data['detection_categories']
    
        images = data['images']
        n_images = len(images)
        if not (options.rename_conf is None and options.rename_cats is None):
            rename_cats = options.rename_cats.split(",")
            if rename_cats[0] == 'all':
                rename_cats = list(category_mapping.keys())
        else:
            rename_cats = []
        if len(options.num_threads) > 0:
            num_threads = int(options.num_threads)
        else:
            num_threads = 1
        print(num_threads)
        if options.xmp_gui is None:
            func = partial(update_xmp_metadata, categories, options, rename_cats, n_images)
            with Pool(num_threads) as p:
                with tqdm(total=n_images) as pbar:
                    for i, _ in enumerate(p.imap_unordered(func, images)):
                        pbar.update()
        else:
            func = partial(update_xmp_metadata, categories, options, rename_cats, n_images)
            with ThreadPool(num_threads) as p:
                p.map(func, images)
            s = 'Successfully processed {} images'.format(n_images)
            print(s)
            write_status(options,s)
        
    except Exception as e:
        
        print('Error processing input data: {}'.format(str(e)))
        traceback.print_exc()
        if options.xmp_gui is not None:
            tkinter.messagebox.showerror(title='Error', 
                                         message='Make Sure you selected the proper image folder and JSON files')
        sys.exit()


def start_input_processing(options):
    
    t = threading.Thread(target=lambda: process_input_data(options))
    t.start()


def browse_folder(options,folder_path_var):
    
    filename = tkinter.filedialog.askdirectory()
    options.image_folder = r'{0}'.format(filename)
    folder_path_var.set(filename)


def browse_file(options,file_path_var):
    
    filename = tkinter.filedialog.askopenfilename()
    options.input_file = r'{0}'.format(filename)
    file_path_var.set(filename)


def create_gui(options):
    
    root = tkinter.Tk()
    root.resizable(False, False)
    root.configure(background='white')
    root.title('DigiKam Integration')
    
    group = tkinter.LabelFrame(root, padx=5, pady=5)
    group.configure(background = 'white')
    group.pack(padx=10, pady=10, fill='both', expand='yes')

    canvas = tkinter.Canvas(group, width = 800, height = 150)    
    canvas.configure(background = 'white')
    canvas.pack()      
    img1 = tkinter.PhotoImage(file='images/aiforearth.png')      
    canvas.create_image(0,0, anchor=tkinter.NW, image=img1)
    img2 = tkinter.PhotoImage(file='images/bg.png')      
    canvas.create_image(0,20, anchor=tkinter.NW, image=img2)

    frame = tkinter.Frame(root)
    frame.configure(background='white')
    frame.pack()
    
    l1 = tkinter.Label(frame, text='Folder containing images')
    l1.configure(background='white')
    l1.grid(row=0, column=0)
    
    folder_path_var = tkinter.StringVar()
    
    e1 = tkinter.Entry(frame, width=50, textvariable=folder_path_var, highlightthickness=1) 
    e1.configure(highlightbackground='grey', highlightcolor='grey')
    e1.grid(row=0, column=2)
    
    b1 = tkinter.Button(frame, text='Browse', fg='blue', command=lambda: browse_folder(options,folder_path_var))
    b1.grid(row=0, column=5, padx=10)
    
    l2 = tkinter.Label(frame, text='Path to MegaDetector output .json file') 
    l2.configure(background='white')
    l2.grid(row=1, column=0)
    
    file_path_var = tkinter.StringVar()
    
    e2 = tkinter.Entry(frame, width=50, textvariable=file_path_var, highlightthickness=1)
    e2.configure(highlightbackground='grey', highlightcolor='grey')    
    e2.grid(row=1, column=2)
    
    b2 = tkinter.Button(frame, text='Browse', fg='blue', command=lambda: browse_file(options,file_path_var))
    b2.grid(row=1, column=5, padx=10)

    l6 = tkinter.Label(frame, text='Minimum confidence to consider a category') 
    l6.configure(background='white')
    l6.grid(row=2, column=0)
    
    textarea_min_threshold = tkinter.Entry(frame, width=50, highlightthickness=1)
    textarea_min_threshold.configure(highlightbackground='grey', highlightcolor='grey')
    textarea_min_threshold.grid(row=2, column=2)
    
    l3 = tkinter.Label(frame, text='Prefix to remove from image paths (optional)') 
    l3.configure(background='white')
    l3.grid(row=3, column=0)
    
    textarea_remove_path = tkinter.Entry(frame, width=50, highlightthickness=1)
    textarea_remove_path.configure(highlightbackground='grey', highlightcolor='grey')
    textarea_remove_path.grid(row=3, column=2)

    l4 = tkinter.Label(frame, text='Confidence level to move images requires manual check (optional)') 
    l4.configure(background='white')
    l4.grid(row=4, column=0)

    textarea_rename_conf = tkinter.Entry(frame, width=50, highlightthickness=1)
    textarea_rename_conf.configure(highlightbackground='grey', highlightcolor='grey')
    textarea_rename_conf.grid(row=4, column=2)


    l5 = tkinter.Label(frame, text='Categories to check for the confidence (optional)') 
    l5.configure(background='white')
    l5.grid(row=5, column=0)

    textarea_rename_cats = tkinter.Entry(frame, width=50, highlightthickness=1)
    textarea_rename_cats.configure(highlightbackground='grey', highlightcolor='grey')
    textarea_rename_cats.grid(row=5, column=2)

    l6 = tkinter.Label(frame, text='Number of threads to run (optional)') 
    l6.configure(background='white')
    l6.grid(row=6, column=0)

    textarea_num_threads = tkinter.Entry(frame, width=50, highlightthickness=1)
    textarea_num_threads.configure(highlightbackground='grey', highlightcolor='grey')
    textarea_num_threads.grid(row=6, column=2)
    
    sb = tkinter.Button(frame, text='Submit', fg='black',
                command=lambda: start_input_processing(options), padx=10)
    sb.grid(row=7, column=2, padx=10, pady=10)

    style = tkinter.ttk.Style(root)
    style.layout('text.Horizontal.Tprogress_bar',
                [('Horizontal.progress_bar.trough',
                {'children': [('Horizontal.progress_bar.pbar',
                                {'side': 'left', 'sticky': 'ns'})],
                    'sticky': 'nswe'}),
                ('Horizontal.progress_bar.label', {'sticky': ''})])    
    style.configure('text.Horizontal.Tprogress_bar', text='0 %')

    progress_bar = tkinter.ttk.Progressbar(root, style='text.Horizontal.Tprogress_bar', length=700,
                                maximum=100, value=0, mode='determinate')
    progress_bar.pack(pady=10)

    group2 = tkinter.LabelFrame(root, text='Status', padx=5, pady=5)
    group2.pack(padx=10, pady=10, fill='both', expand='yes')

    textarea_status = tkinter.Text(group2, height=10, width=100)
    textarea_status.configure(state="disabled")
    textarea_status.pack()
    
    options.xmp_gui = xmp_gui()
    options.xmp_gui.root = root
    options.xmp_gui.textarea_min_threshold = textarea_min_threshold
    options.xmp_gui.textarea_remove_path = textarea_remove_path
    options.xmp_gui.textarea_rename_conf = textarea_rename_conf
    options.xmp_gui.textarea_rename_cats = textarea_rename_cats
    options.xmp_gui.textarea_num_threads = textarea_num_threads
    options.xmp_gui.textarea_status = textarea_status
    options.xmp_gui.progress_bar = progress_bar
    options.xmp_gui.style = style
    
    root.mainloop()
    
    
#%% Interactive/test driver

if False:
    
    #%%
    
    options = xmp_integration_options()
    options.input_file = r"C:\temp\demo_images\ssmini_xmp_test_orig\ssmini.mdv4.json"
    options.image_folder = r"C:\temp\demo_images\ssmini_xmp_test"
    options.remove_path = 'my_images/'
    process_input_data(options)
    
    
#%% Command-line driver

def args_to_object(args,obj):
    """
    Copy all fields from the argparse table "args" to the object "obj"
    """
    for n, v in inspect.getmembers(args):
        if not n.startswith('_'):
            setattr(obj, n, v)


def main(): # noqa
    
    parser = argparse.ArgumentParser()
    parser.add_argument('--input_file', help = 'Path to the MegaDetector .json file', default=None)
    parser.add_argument('--image_folder', help = 'Path to the folder containing images', default=None)
    parser.add_argument('--min_threshold', help = 'Minimum detection confidence that will be treated as a detection event', default=None)
    parser.add_argument('--remove_path', help = 'Prefix to remove from image paths in the .json file', default=None)
    parser.add_argument('--rename_conf', help = 'Below this confidence level, images will be renamed for manual check', default=None)
    parser.add_argument('--rename_cat', help = 'Category (or comma-delimited categories) to apply renaming behavior to', default=None)
    parser.add_argument('--num_threads', help = 'Number of threads to use for image processing', default=1)
    parser.add_argument('--gui', help = 'Run in GUI mode', action='store_true')
    
    options = xmp_integration_options()
    args = parser.parse_args()
    args_to_object(args,options)
    
    if options.gui:
        assert options.input_file is None, 'Command-line argument specified in GUI mode'
        assert options.image_folder is None, 'Command-line argument specified in GUI mode'
        assert options.min_threshold is None, 'Command-line argument specified in GUI mode'
        assert options.remove_path is None, 'Command-line argument specified in GUI mode'
        assert options.rename_conf is None, 'Command-line argument specified in GUI mode'
        assert options.rename_cat is None, 'Command-line argument specified in GUI mode'
        assert options.num_threads == 1, 'Command-line argument specified in GUI mode'
        create_gui(options)    
    else:
        process_input_data(options)

if __name__ == '__main__':
    
    main()



"""

create_links_to_md_results_files.py

One-off script to populate the columns in the camera trap data .csv file that point to MD results.

"""

#%% Imports and constants

import os

import pandas as pd

input_csv_file = r'g:\temp\lila_camera_trap_datasets_no_md_results.csv'
output_csv_file = r'g:\temp\lila_camera_trap_datasets.csv'

md_results_local_folder = r'g:\temp\lila-md-results'
md_base_url = 'https://lila.science/public/lila-md-results/'
assert md_base_url.endswith('/')

# No RDE files for datasets with no location information
datasets_without_location_info = ('ena24','missouri-camera-traps')

md_results_column_names = ['mdv4_results_raw','mdv5a_results_raw','mdv5b_results_raw','md_results_with_rde']

validate_urls = False


#%% Read input data

df = pd.read_csv(input_csv_file)
for s in md_results_column_names:
    df[s] = ''


#%% Find matching files locally, and create URLs

local_files = os.listdir(md_results_local_folder)
local_files = [fn for fn in local_files if fn.endswith('.zip')]

# i_row = 0; row = df.iloc[i_row]
for i_row,row in df.iterrows():

    if not isinstance(row['name'],str):
        continue

    dataset_shortname = row['short_name']
    matching_files = [fn for fn in local_files if dataset_shortname in fn]

    # No RDE files for datasets with no location information
    if dataset_shortname in datasets_without_location_info:
        assert len(matching_files) == 2
        mdv5a_files = [fn for fn in matching_files if 'mdv5a' in fn]
        mdv5b_files = [fn for fn in matching_files if 'mdv5b' in fn]
        assert len(mdv5a_files) == 1 and len(mdv5b_files) == 1
        df.loc[i_row,'mdv5a_results_raw'] = md_base_url + mdv5a_files[0]
        df.loc[i_row,'mdv5b_results_raw'] = md_base_url + mdv5b_files[0]
    else:
        # Exclude single-season files for snapshot-serengeti
        if dataset_shortname == 'snapshot-serengeti':
            matching_files = [fn for fn in matching_files if '_S' not in fn]
            assert len(matching_files) == 2
            assert all(['mdv4' in fn for fn in matching_files])
            rde_files = [fn for fn in matching_files if 'rde' in fn]
            raw_files = [fn for fn in matching_files if 'rde' not in fn]
            assert len(rde_files) == 1 and len(raw_files) == 1
            df.loc[i_row,'mdv4_results_raw'] = md_base_url + raw_files[0]
            df.loc[i_row,'md_results_with_rde'] = md_base_url + rde_files[0]
        else:
            assert len(matching_files) == 3
            mdv5a_files = [fn for fn in matching_files if 'mdv5a' in fn and 'rde' not in fn]
            mdv5b_files = [fn for fn in matching_files if 'mdv5b' in fn and 'rde' not in fn]
            rde_files = [fn for fn in matching_files if 'rde' in fn]
            assert len(mdv5a_files) == 1 and len(mdv5b_files) == 1 and len(rde_files) == 1
            df.loc[i_row,'mdv5a_results_raw'] = md_base_url + mdv5a_files[0]
            df.loc[i_row,'mdv5b_results_raw'] = md_base_url + mdv5b_files[0]
            df.loc[i_row,'md_results_with_rde'] = md_base_url + rde_files[0]

    print('Found {} matching files for {}'.format(len(matching_files),dataset_shortname))

# ...for each row


#%% Validate URLs

if validate_urls:

    from megadetector.utils.url_utils import test_urls

    urls = set()

    for i_row,row in df.iterrows():
        for column_name in md_results_column_names:
            if len(row[column_name]) > 0:
                assert row[column_name] not in urls
                urls.add(row[column_name])

    test_urls(urls,error_on_failure=True)

    print('Validated {} URLs'.format(len(urls)))


#%% Write new .csv file

df.to_csv(output_csv_file,header=True,index=False)
